"""Module with some of the common functions, like file and image operations"""

import os
import gi
import random
import shutil
import imageio
from pathlib import Path
from typing import List
from PIL import Image

gi.require_version("Gtk", "3.0")
from gi.repository import Gtk, GdkPixbuf, Gdk, GLib

from waypaper.options import IMAGE_EXTENSIONS, BACKEND_OPTIONS, VIDEO_EXTENSIONS


def has_image_extension(file_path: str, backend: str) -> bool:
    """Check if the file has image extension"""
    image_extensions = IMAGE_EXTENSIONS[backend]
    ext = os.path.splitext(file_path)[1].lower()
    return ext in image_extensions


def get_image_paths(backend: str,
                    folder_list: list[Path],
                    include_subfolders: bool = False,
                    include_all_subfolders: bool = False,
                    include_hidden: bool = False,
                    only_gifs: bool = False) -> list[str]:
    """Get a list of file paths depending on the filters that were requested."""

    image_path_list: list[str] = []
    for folder in folder_list:
        for root, directories, files in os.walk(folder, followlinks=True):

            # Remove hidden files from consideration:
            for directory in directories:
                if directory.startswith('.') and not include_hidden:
                    directories.remove(directory)

            # Remove subfolders from consideration:
            if not include_subfolders and str(root) != str(folder):
                continue

            # Remove deep folders from consideration:
            if not include_all_subfolders and str(root) != str(folder):
                current_depth = root.count(os.path.sep) - str(folder).count(os.path.sep)
                if current_depth > 1:
                    continue

            # Remove files that are not images from consideration:
            for image_name in files:
                if image_name.startswith('.') and not include_hidden:
                    continue
                if not has_image_extension(image_name, backend):
                    continue
                if not image_name.casefold().endswith('.gif') and only_gifs:
                    continue
                image_path_list.append(os.path.join(root, image_name))
    return image_path_list


def get_image_name(full_path: str, base_folder_list: list[Path], include_path: bool) ->  str:
    """Get image name that may or may not include parent folders"""
    full_path = Path(full_path).resolve()

    # If path is not required, just return file name:
    if not include_path:
        return full_path.name

    # Otherwise, find from which folder file comes from and append this folder:
    for base_folder in base_folder_list:
        base_folder = Path(base_folder).resolve()
        if not full_path.is_relative_to(base_folder):
            continue
        common_folder = base_folder.name
        return str(Path(common_folder, full_path.relative_to(base_folder)))


def get_random_file(backend: str,
                    folder_list: list[Path],
                    include_subfolders: bool,
                    include_all_subfolders: bool,
                    cache_dir: Path,
                    include_hidden: bool = False) -> str | None:
    """Pick a random file from the folder and update cache"""
    try:
        # Get all image paths from the folder:
        image_paths = get_image_paths(backend, folder_list, include_subfolders, include_all_subfolders,
                                      include_hidden, only_gifs=False)

        # Read cache file with already used images:
        cache_file = cache_dir / "used_wallpapers.txt"
        if cache_file.exists():
            with cache_file.open('r') as file:
                used_images = [line.strip() for line in file.readlines()]
        # Create it if the file does not exists:
        else:
            cache_file.touch()
            used_images = []

        # Pick a random image from unused images:
        remaining_images = list(filter(lambda img: img not in set(used_images), image_paths))
        if remaining_images:
            random_image = random.choice(remaining_images)
            used_images.append(random_image)
        else:
            random_image = random.choice(image_paths)
            used_images = [random_image]

        # Write the cache file:
        with cache_file.open('w') as file:
            for img in used_images:
                file.write(img + '\n')

        return random_image

    except Exception as e:
        print(f"Error getting random image: {e}")
        return None


def check_installed_backends() -> List[str]:
    """Check which backends are installed in the system"""
    installed_backends = ["none"]
    for backend in BACKEND_OPTIONS:
        if backend == "none":
            continue
        if backend == "wallutils":
            binary_name = "setwallpaper"
        else:
            binary_name = backend
        is_installed = bool(shutil.which(binary_name))
        if is_installed:
            installed_backends.append(backend)
    return installed_backends


def cache_image(image_path: str, cache_dir: Path) -> None:
    """Create small copies of images using various libraries depending on the file type"""
    ext = os.path.splitext(image_path)[1].lower()
    cache_file = cache_dir / Path(os.path.basename(image_path))
    width = 240
    try:
        # If it's a video, extract the first frame:
        if ext in VIDEO_EXTENSIONS:
            reader = imageio.get_reader(image_path)
            first_frame = reader.get_data(0)
            # Convert the numpy array to a PIL image:
            pil_image = Image.fromarray(first_frame)
            aspect_ratio = pil_image.height / pil_image.width
            new_height = int(width * aspect_ratio)
            resized_image = pil_image.resize((width, new_height))
            resized_image.save(str(cache_file), "JPEG")
            return

        # If it's an image, create preview depending on the filetype
        if ext == ".webp":
            img = Image.open(image_path)
            data = img.tobytes()
            img_width, img_height = img.size
            pixbuf = GdkPixbuf.Pixbuf.new_from_data(data, GdkPixbuf.Colorspace.RGB, False, 8, img_width, img_height, img_width * 3)
        else:
            pixbuf = GdkPixbuf.Pixbuf.new_from_file(str(image_path))
        aspect_ratio = pixbuf.get_width() / pixbuf.get_height()
        height = int(width / aspect_ratio)
        scaled_pixbuf = pixbuf.scale_simple(width, height, GdkPixbuf.InterpType.BILINEAR)
        scaled_pixbuf.savev(str(cache_file), "jpeg", [], [])

    # If image processing failed, create a black placeholder:
    except Exception as e:
        print(f"Could not generate preview for {os.path.basename(image_path)}")
        print(e)
        black_pixbuf = GdkPixbuf.Pixbuf.new(GdkPixbuf.Colorspace.RGB, True, 8, width, width*9/16)
        black_pixbuf.fill(0x0)
        black_pixbuf.savev(str(cache_file), "jpeg", [], [])



"""Module that contains lists of possible options used in the application"""

from typing import List, Dict
from screeninfo import get_monitors

BACKEND_OPTIONS: List[str] = ["none", "swaybg", "swww", "feh", "wallutils", "hyprpaper", "mpvpaper"]
FILL_OPTIONS: List[str] = ["fill", "stretch", "fit", "center", "tile"]
SORT_OPTIONS: List[str] = ["name", "namerev", "date", "daterev"]
SORT_DISPLAYS: Dict[str, str] = {"name": "Name ↓", "namerev": "Name ↑", "date": "Date ↓", "daterev": "Date ↑"}
MONITOR_OPTIONS: List[str] = [m.name for m in get_monitors()]

VIDEO_EXTENSIONS: List[str] = ['.webm', '.mkv', '.flv', '.vob', '.ogv', '.ogg', '.rrc', '.gifv', '.mng', '.mov',
                               '.avi', '.qt', '.wmv', '.yuv', '.rm', '.asf', '.amv', '.mp4', '.m4p', '.m4v',
                               '.mpg', '.mp2', '.mpeg', '.mpe', '.mpv', '.m4v', '.svi', '.3gp', '.3g2', '.mxf',
                               '.roq', '.nsv', '.flv', '.f4v', '.f4p', '.f4a', '.f4b', '.mod' ]

IMAGE_EXTENSIONS: Dict[str, List[str]] = {
        BACKEND_OPTIONS[0]: ['.gif', '.jpg', '.jpeg', '.png', '.webp', '.bmp', '.pnm', '.tiff'],
        BACKEND_OPTIONS[1]: ['.gif', '.jpg', '.jpeg', '.png'],
        BACKEND_OPTIONS[2]: ['.gif', '.jpg', '.jpeg', '.png', '.webp', '.bmp', '.pnm', '.tiff'],
        BACKEND_OPTIONS[3]: ['.gif', '.jpg', '.jpeg', '.png', '.bmp', '.pnm', '.tiff'],
        BACKEND_OPTIONS[4]: ['.gif', '.jpg', '.jpeg', '.png'],
        BACKEND_OPTIONS[5]: ['.jpg', '.jpeg', '.png', '.webp'],
        BACKEND_OPTIONS[6]: ['.gif', '.jpg', '.jpeg', '.png', '.webp', '.bmp', '.pnm', '.tiff', '.avif'] + VIDEO_EXTENSIONS,
        }

SWWW_TRANSITION_TYPES: List[str] = ["any", "none", "simple", "fade", "wipe",  "left", "right", "top",
                                "bottom", "wave", "grow", "center", "outer", "random"]

TIMERS: Dict[str, int] = {"30 sec": 30, "1 min": 60, "2 min": 120, "5 min": 300, "10 min": 600, "30 min": 1800, "1 hour": 3600,
          "2 hours": 7200, "6 hours": 21600, "12 hours": 43200, "1 day": 86400, "1 week": 604800}



"""Main module that either runs cli commands or starts GUI"""

import argparse
import sys
import time
import json
import pathlib
import threading

from waypaper.app import App
from waypaper.changer import change_wallpaper
from waypaper.common import get_random_file
from waypaper.config import Config
from waypaper.options import BACKEND_OPTIONS, FILL_OPTIONS, MONITOR_OPTIONS
from waypaper.translations import load_language


__version__ = "2.6"

# Get application settings and language package:
cf = Config()
cf.read()
cf.read_state()
txt = load_language(cf.lang)

# Define command line argument parser and parse user arguments:
parser = argparse.ArgumentParser(
    prog="waypaper", description=txt.msg_desc, epilog=txt.msg_info
)
parser.add_argument("-v", "--version", help=txt.msg_arg_help, action="store_true")
parser.add_argument("--restore", help=txt.msg_arg_rest, action="store_true")
parser.add_argument("--random", help=txt.msg_arg_rand, action="store_true")
parser.add_argument("--fill", help=txt.msg_arg_fill, choices=FILL_OPTIONS)
parser.add_argument("--wallpaper", help=txt.msg_arg_wall)
parser.add_argument("--folder", help=txt.msg_arg_folder, nargs="+", default = [])
parser.add_argument("--state-file", help=txt.msg_arg_statefile)
parser.add_argument("--backend", help=txt.msg_arg_back, choices=BACKEND_OPTIONS)
parser.add_argument("--list", help=txt.msg_arg_list, action='store_true')
parser.add_argument("--monitor", help=txt.msg_arg_monitor, choices=MONITOR_OPTIONS)
parser.add_argument("--no-post-command", help=txt.msg_arg_post, action='store_true')
args = parser.parse_args()


def run():
    """Read user arguments and either run GUI app or perform requested action"""

    # Read user arguments, and update things if alternative state file was provided:
    cf.read_parameters_from_user_arguments(args)
    if args.state_file:
        cf.read_state()
        cf.read_parameters_from_user_arguments(args)
    cf.check_validity()

    # Set monitor and wallpaper from user arguments:
    if args.monitor:
        monitor = args.monitor

        # Set wallpaper from user arguments if provided:
        if args.wallpaper:
            wallpaper = pathlib.Path(args.wallpaper).expanduser()

        # Otherwise set a random wallpaper:
        else:
            wallpaper_str = get_random_file(cf.backend, cf.image_folder_list, cf.include_subfolders,
                                            cf.include_all_subfolders, cf.cache_dir, cf.show_hidden)
            if wallpaper_str:
                wallpaper = pathlib.Path(wallpaper_str)
            else:
                print("Could not get random wallpaper.")
                sys.exit(0)

        # Launch commands to change wallpaper in a separate thread:
        threading.Thread(target=change_wallpaper, args=(wallpaper, cf, monitor)).start()
        time.sleep(0.1)

        # Save this wallpaper in config and quit:
        cf.selected_wallpaper = wallpaper
        cf.selected_monitor = monitor
        cf.attribute_selected_wallpaper()
        if cf.use_xdg_state:
            cf.save_state_file()
        else:
            cf.save()
        sys.exit(0)

    # Set previous wallpapers or random wallpaper:
    if args.restore or args.random:
        for index, (wallpaper, monitor) in enumerate(zip(cf.wallpapers, cf.monitors)):
            if args.random:
                wallpaper_str = get_random_file(cf.backend, cf.image_folder_list, cf.include_subfolders,
                                                cf.include_all_subfolders, cf.cache_dir, cf.show_hidden)
                if wallpaper_str:
                    wallpaper = pathlib.Path(wallpaper_str)
                    cf.wallpapers[index] = wallpaper

            if cf.wallpapers[index] is None:
                continue

            # Launch commands to change wallpaper in a separate thread:
            threading.Thread(target=change_wallpaper, args=(wallpaper, cf, monitor)).start()
            time.sleep(0.1)

        # Save new wallpapers:
        if cf.use_xdg_state:
            cf.save_state_file()
        else:
            cf.save()
        sys.exit(0)

    # Set wallpaper from user arguments:
    if args.wallpaper:
        monitor = "All"
        wallpaper = pathlib.Path(args.wallpaper).expanduser()
        threading.Thread(target=change_wallpaper, args=(wallpaper, cf, monitor)).start()

        # Save this wallpaper in config and quit:
        cf.selected_wallpaper = wallpaper
        cf.selected_monitor = monitor
        cf.attribute_selected_wallpaper()
        if cf.use_xdg_state:
            cf.save_state_file()
        else:
            cf.save()
        sys.exit(0)

    # Output some information in json format:
    if args.list:
        info = list(map(lambda x: {"monitor": x[0], "wallpaper": str(x[1]), "backend": cf.backend}, zip(cf.monitors, cf.wallpapers)))
        print(json.dumps(info))
        sys.exit(0)

    # Print the version and quit:
    if args.version:
        print(f"waypaper v.{__version__}")
        sys.exit(0)

    # Start GUI:
    app = App(txt, cf)
    app.run()


if __name__ == "__main__":
    run()



"""Module that runs GUI app"""

import threading
import subprocess
import os
import gi
import shutil
import imageio
import screeninfo
from pathlib import Path

from waypaper.changer import change_wallpaper
from waypaper.config import Config
from waypaper.common import get_image_paths, get_image_name, get_random_file, cache_image
from waypaper.options import FILL_OPTIONS, SORT_OPTIONS, SORT_DISPLAYS, VIDEO_EXTENSIONS , SWWW_TRANSITION_TYPES
from waypaper.translations import Chinese, English, French, German, Polish, Russian, Belarusian, Spanish

gi.require_version("Gtk", "3.0")
from gi.repository import Gtk, GdkPixbuf, Gdk, GLib


class App(Gtk.Window):
    """Main application class that controls GUI"""

    def __init__(self, txt: Chinese|English|French|German|Polish|Russian|Belarusian|Spanish, cf: Config) -> None:
        super().__init__(title="Waypaper")
        self.cf = cf
        self.txt = txt
        self.check_backends()
        self.set_default_size(820, 600)
        self.connect("delete-event", Gtk.main_quit)
        self.selected_index = 0
        self.highlighted_image_row = 0
        self.is_enering_text = False
        self.number_of_resize = 0
        self.init_ui()
        self.main_box.grab_focus()

        # Start the image processing in a separate thread:
        threading.Thread(target=self.process_images).start()


    def init_ui(self) -> None:
        """Initialize the UI elements of the application"""

        # Create a vertical box for general app layout:
        self.main_box = Gtk.VBox(spacing=10)
        self.add(self.main_box)

        # Load and apply CSS
        css_provider = Gtk.CssProvider()
        css = b"""
        .highlighted-button {
            border: 1px solid @theme_selected_bg_color;
        }
        """
        css_provider.load_from_data(css)

        # Apply CSS to the default screen
        screen = Gdk.Screen.get_default()
        context = Gtk.StyleContext()
        context.add_provider_for_screen(screen, css_provider, Gtk.STYLE_PROVIDER_PRIORITY_APPLICATION)

        # TOP MENU

        # Create a box to contain the top row of items:
        self.top_button_box = Gtk.Box(orientation=Gtk.Orientation.HORIZONTAL, spacing=60)
        self.top_button_box.set_margin_top(20)
        self.main_box.pack_start(self.top_button_box, False, False, 0)

        # Create alignment container for top menu:
        self.top_row_alignment = Gtk.Alignment(xalign=0.5, yalign=0.0, xscale=0.5, yscale=0.5)
        self.top_button_box.pack_start(self.top_row_alignment, True, False, 0)

        # Create a button to open folder dialog:
        self.choose_folder_button = Gtk.Button(label=self.txt.msg_changefolder)
        self.choose_folder_button.connect("clicked", self.on_choose_folder_clicked)

        # Create a search entry:
        self.search_entry = Gtk.Entry()
        self.search_entry.set_placeholder_text(self.txt.msg_search)
        self.search_entry.connect("changed", self.on_search_entry_changed)
        self.search_entry.connect("focus-in-event", self.on_focus_in)
        self.search_entry.connect("focus-out-event", self.on_focus_out)

        # Create a clear button:
        self.clear_button = Gtk.Button(label=self.txt.msg_clear)
        self.clear_button.connect("clicked", self.on_clear_button)

        # Create the options menu button:
        self.options_button = Gtk.Button(label=self.txt.msg_options)
        self.options_button.connect("clicked", self.on_options_button_clicked)

        # Create a sort option dropdown menu:
        self.sort_combo = Gtk.ComboBoxText()
        for option in SORT_OPTIONS:
            self.sort_combo.append_text(SORT_DISPLAYS[option])
        active_num = SORT_OPTIONS.index(self.cf.sort_option)
        self.sort_combo.set_active(active_num)
        self.sort_combo.connect("changed", self.on_sort_option_changed)
        self.sort_combo.set_tooltip_text(self.txt.tip_sorting)

        # Create refresh button:
        self.refresh_button = Gtk.Button(label=self.txt.msg_refresh)
        self.refresh_button.connect("clicked", self.on_refresh_clicked)
        self.refresh_button.set_tooltip_text(self.txt.tip_refresh)

        # Create random button:
        self.random_button = Gtk.Button(label=self.txt.msg_random)
        self.random_button.connect("clicked", self.on_random_clicked)
        self.random_button.set_tooltip_text(self.txt.tip_random)

        # Create exit button:
        self.exit_button = Gtk.Button(label=self.txt.msg_exit)
        self.exit_button.connect("clicked", self.on_exit_clicked)
        self.exit_button.set_tooltip_text(self.txt.tip_exit)

        # Add all top objects to the container:
        self.top_container = Gtk.Box(orientation=Gtk.Orientation.HORIZONTAL, spacing=10)
        self.top_container.pack_start(self.choose_folder_button, False, False, 0)
        self.top_container.pack_start(self.search_entry, expand=False, fill=False, padding=0)
        self.top_container.pack_start(self.clear_button, expand=False, fill=False, padding=0)
        self.top_container.pack_start(self.sort_combo, expand=False, fill=False, padding=0)
        self.top_container.pack_start(self.refresh_button, expand=False, fill=False, padding=0)
        self.top_container.pack_start(self.random_button, expand=False, fill=False, padding=0)
        self.top_container.pack_start(self.options_button, expand=False, fill=False, padding=0)
        self.top_container.pack_start(self.exit_button, expand=False, fill=False, padding=0)
        self.top_row_alignment.add(self.top_container)

        # MIDDLE GRID

        # Create a scrolled window for the grid of images:
        self.scrolled_window = Gtk.ScrolledWindow()
        self.scrolled_window.set_policy(Gtk.PolicyType.AUTOMATIC, Gtk.PolicyType.AUTOMATIC)

        # Create a box to center the grid:
        self.center_box = Gtk.Box(orientation=Gtk.Orientation.VERTICAL)
        self.center_box.set_valign(Gtk.Align.CENTER)
        self.center_box.set_halign(Gtk.Align.CENTER)
        self.main_box.add(self.scrolled_window)

        # Create a grid layout for images:
        self.grid = Gtk.Grid()
        self.grid.set_row_spacing(0)
        self.grid.set_column_spacing(0)
        self.center_box.add(self.grid)
        self.scrolled_window.add(self.center_box)

        # BACKEND MENU

        # Create a backend dropdown menu:
        self.backend_option_combo = Gtk.ComboBoxText()
        for backend in self.cf.installed_backends:
            self.backend_option_combo.append_text(backend)

        # Set as active line the backend from config, if it is installed:
        active_num = 0
        if self.cf.backend in self.cf.installed_backends:
            active_num = self.cf.installed_backends.index(self.cf.backend)
        self.backend_option_combo.set_active(active_num)
        self.backend_option_combo.connect("changed", self.on_backend_option_changed)
        self.backend_option_combo.set_tooltip_text(self.txt.tip_backend)

        # Create a fill option dropdown menu:
        self.fill_option_combo = Gtk.ComboBoxText()
        for option in FILL_OPTIONS:
            capitalized_option = option[0].upper() + option[1:]
            self.fill_option_combo.append_text(capitalized_option)
        if self.cf.fill_option in FILL_OPTIONS:
            active_fill_option_index = FILL_OPTIONS.index(self.cf.fill_option)
            self.fill_option_combo.set_active(active_fill_option_index)
        else:
            self.fill_option_combo.set_active(0)
        self.fill_option_combo.connect("changed", self.on_fill_option_changed)
        self.fill_option_combo.set_tooltip_text(self.txt.tip_fill)

        # Create a color picker:
        self.color_picker_button = Gtk.ColorButton()
        self.color_picker_button.set_use_alpha(True)
        rgba_color = Gdk.RGBA()
        rgba_color.parse(self.cf.color)
        self.color_picker_button.set_rgba(rgba_color)
        self.color_picker_button.connect("color-set", self.on_color_set)
        self.color_picker_button.set_tooltip_text(self.txt.tip_color)

        # Create mpv stop button:
        self.mpv_stop_button = Gtk.Button(label=self.txt.msg_stop)
        self.mpv_stop_button.connect("clicked", self.on_mpv_stop_button_clicked)
        self.mpv_stop_button.set_tooltip_text(self.txt.tip_mpv_stop)

        # Create mpv pause button:
        self.mpv_pause_button = Gtk.Button(label=self.txt.msg_pause)
        self.mpv_pause_button.connect("clicked", self.on_mpv_pause_button_clicked)
        self.mpv_pause_button.set_tooltip_text(self.txt.tip_mpv_pause)

        # Create mpv sound toggle:
        self.mpv_sound_toggle = Gtk.ToggleButton(label=self.txt.msg_sound)
        self.mpv_sound_toggle.set_active(self.cf.mpvpaper_sound)
        self.mpv_sound_toggle.connect("toggled", self.on_mpv_sound_toggled)
        self.mpv_sound_toggle.set_tooltip_text(self.txt.tip_mpv_sound)

        # Create a box to contain the bottom row of buttons:
        self.bottom_button_box = Gtk.Box(orientation=Gtk.Orientation.HORIZONTAL, spacing=60)
        self.bottom_button_box.set_margin_bottom(15)
        self.main_box.pack_end(self.bottom_button_box, False, False, 0)

        # Create a box to contain the loading label:
        self.bottom_loading_box = Gtk.HBox(spacing=0)
        self.bottom_loading_box.set_margin_bottom(0)
        self.main_box.pack_end(self.bottom_loading_box, False, False, 0)

        # Create alignment container for bottom menu:
        self.button_row_alignment = Gtk.Alignment(xalign=0.5, yalign=0.0, xscale=0.5, yscale=0.5)
        self.bottom_button_box.pack_start(self.button_row_alignment, True, False, 0)

        # Create a monitor option dropdown menu:
        self.monitor_option_combo = Gtk.ComboBoxText()

        # Create a horizontal box for display backend options:
        self.options_box = Gtk.HBox(spacing=10)
        self.options_box.pack_start(self.backend_option_combo, False, False, 0)
        self.button_row_alignment.add(self.options_box)

        # Create a transition type dropdown menu for swww
        self.swww_transitions_options = Gtk.ComboBoxText()

        #  Get angle for animation
        self.swww_angle_entry = Gtk.Entry()
        self.swww_angle_entry.set_width_chars(5)
        self.swww_angle_entry.set_placeholder_text("angle")
        self.swww_angle_entry.connect("focus-in-event", self.on_focus_in)
        self.swww_angle_entry.connect("focus-out-event", self.on_focus_out)

        #  Get steps for animation
        self.swww_steps_entry = Gtk.Entry()
        self.swww_steps_entry.set_width_chars(5)
        self.swww_steps_entry.set_placeholder_text("steps")
        self.swww_steps_entry.connect("focus-in-event", self.on_focus_in)
        self.swww_steps_entry.connect("focus-out-event", self.on_focus_out)

        #  Get duration for animation
        self.swww_duration_entry = Gtk.Entry()
        self.swww_duration_entry.set_width_chars(7)
        self.swww_duration_entry.set_placeholder_text("duration")
        self.swww_duration_entry.connect("focus-in-event", self.on_focus_in)
        self.swww_duration_entry.connect("focus-out-event", self.on_focus_out)

        #  Get fps for animation
        self.swww_fps_entry = Gtk.Entry()
        self.swww_fps_entry.set_width_chars(5)
        self.swww_fps_entry.set_placeholder_text("fps")
        self.swww_fps_entry.connect("focus-in-event", self.on_focus_in)
        self.swww_fps_entry.connect("focus-out-event", self.on_focus_out)

        # Add different buttons depending on backend:
        self.monitor_option_display()
        self.mpv_options_display()
        self.fill_option_display()
        self.color_picker_display()
        self.swww_options_display()

        # Connect the key press events to various actions:
        self.connect("key-press-event", self.on_key_pressed)

        self.connect("button-press-event", self.on_window_clicked)

        # Connect window resizing events to change the number of columns.
        # self.connect("size-allocate", self.on_window_resize)

        self.show_all()


    def create_options_menu(self) -> None:
        """Create a GTK menu with some options of the application"""
        self.menu = Gtk.Menu()

        # Create gifs toggle:
        self.filter_gifs_checkbox = Gtk.CheckMenuItem(label=self.txt.msg_gifs)
        self.filter_gifs_checkbox.set_active(self.cf.show_gifs_only)
        self.filter_gifs_checkbox.connect("toggled", self.on_filter_gifs_toggled)
        self.menu.append(self.filter_gifs_checkbox)

        # Create subfolder toggle:
        self.include_subfolders_checkbox = Gtk.CheckMenuItem(label=self.txt.msg_subfolders)
        self.include_subfolders_checkbox.set_active(self.cf.include_subfolders)
        self.include_subfolders_checkbox.connect("toggled", self.on_include_subfolders_toggled)
        self.menu.append(self.include_subfolders_checkbox)

        # Create all subfolder toggle:
        self.include_all_subfolders_checkbox = Gtk.CheckMenuItem(label=self.txt.msg_all_subfolders)
        self.include_all_subfolders_checkbox.set_active(self.cf.include_all_subfolders)
        self.include_all_subfolders_checkbox.connect("toggled", self.on_include_all_subfolders_toggled)
        self.menu.append(self.include_all_subfolders_checkbox)

        # Create hidden toggle:
        self.include_hidden_checkbox = Gtk.CheckMenuItem(label=self.txt.msg_hidden)
        self.include_hidden_checkbox.set_active(self.cf.show_hidden)
        self.include_hidden_checkbox.connect("toggled", self.on_hidden_files_toggled)
        self.menu.append(self.include_hidden_checkbox)

        # Create show folder path toggle:
        self.show_path_in_tooltip_checkbox = Gtk.CheckMenuItem(label=self.txt.msg_show_path_in_tooltip)
        self.show_path_in_tooltip_checkbox.set_active(self.cf.show_path_in_tooltip)
        self.show_path_in_tooltip_checkbox.connect("toggled", self.on_show_path_in_tooltip_toggled)
        self.menu.append(self.show_path_in_tooltip_checkbox)

        # Create zen mode toggle:
        self.zen_mode_checkbox = Gtk.CheckMenuItem(label=self.txt.msg_zen)
        self.zen_mode_checkbox.set_active(self.cf.zen_mode)
        self.zen_mode_checkbox.connect("toggled", self.on_zen_mode_toggled)
        self.menu.append(self.zen_mode_checkbox)

        self.menu.show_all()


    def on_options_button_clicked(self, widget) -> None:
        '''Position the menu at the button and show it'''
        self.create_options_menu()
        self.menu.popup_at_widget(widget, Gdk.Gravity.NORTH, Gdk.Gravity.SOUTH, None)

    def monitor_option_display(self) -> None:
        """Display monitor option if backend is not feh or wallutils"""
        self.options_box.remove(self.monitor_option_combo)
        # Check available monitors:
        monitor_names = ["All"]
        if self.cf.backend in ["feh", "wallutils", "none"]:
            return
        monitor_names.extend([m.name for m in screeninfo.get_monitors()])

        # Create a monitor option dropdown menu:
        self.monitor_option_combo = Gtk.ComboBoxText()
        for monitor in monitor_names:
            self.monitor_option_combo.append_text(monitor)
            self.monitor_option_combo.set_active(0)
            self.monitor_option_combo.connect("changed", self.on_monitor_option_changed)
            self.monitor_option_combo.set_tooltip_text(self.txt.tip_display)

        # Add it to the row of buttons:
        self.options_box.pack_start(self.monitor_option_combo, False, False, 0)

    def swww_options_display(self) -> None:
        """Show swww transition options if backend is swww"""
        self.options_box.remove(self.swww_transitions_options)
        self.options_box.remove(self.swww_angle_entry)
        self.options_box.remove(self.swww_steps_entry)
        self.options_box.remove(self.swww_fps_entry)
        self.options_box.remove(self.swww_duration_entry)

        if self.cf.backend != "swww":
            return

        self.swww_transitions_options = Gtk.ComboBoxText()
        for transitions in SWWW_TRANSITION_TYPES:
            self.swww_transitions_options.append_text(transitions)
        active_transition = 0
        if self.cf.swww_transition_type in SWWW_TRANSITION_TYPES:
            active_transition = SWWW_TRANSITION_TYPES.index(self.cf.swww_transition_type)
            self.swww_transitions_options.set_active(active_transition)
            self.swww_transitions_options.connect("changed", self.on_transition_option_changed)
            self.swww_transitions_options.set_tooltip_text(self.txt.tip_transition)

        self.options_box.pack_end(self.swww_steps_entry, False, False, 0)
        self.options_box.pack_end(self.swww_fps_entry, False, False, 0)
        self.options_box.pack_end(self.swww_angle_entry, False, False, 0)
        self.options_box.pack_end(self.swww_duration_entry, False, False, 0)
        self.options_box.pack_end(self.swww_transitions_options, False, False, 0)

    def swww_options_read(self) -> None:
        """Read swww transition options from the UI if they are valid"""
        if self.cf.backend != "swww":
            return
        angle = self.swww_angle_entry.get_text()
        steps = self.swww_steps_entry.get_text()
        fps = self.swww_fps_entry.get_text()
        duration = self.swww_duration_entry.get_text()
        if angle.isdigit():
            self.cf.swww_transition_angle = angle
        if steps.isdigit():
            self.cf.swww_transition_step = steps
        if fps.isdigit():
            self.cf.swww_transition_fps = fps
        if duration.isdigit():
            self.cf.swww_transition_duration = duration

    def mpv_options_display(self) -> None:
        """Show mpv options if backend is mpvpaper, and remove them for other backends"""
        self.options_box.remove(self.mpv_stop_button)
        self.options_box.remove(self.mpv_pause_button)
        self.options_box.remove(self.mpv_sound_toggle)
        if self.cf.backend == "mpvpaper":
            self.options_box.pack_end(self.mpv_stop_button, False, False, 0)
            self.options_box.pack_end(self.mpv_pause_button, False, False, 0)
            self.options_box.pack_end(self.mpv_sound_toggle, False, False, 0)

    def fill_option_display(self):
        """Display fill option if backend is not hyprpaper"""
        self.options_box.remove(self.fill_option_combo)
        if self.cf.backend not in ['hyprpaper', 'none']:
            self.options_box.pack_end(self.fill_option_combo, False, False, 0)

    def color_picker_display(self):
        """Display color option if backend is not hyprpaper"""
        self.options_box.remove(self.color_picker_button)
        if self.cf.backend not in ['hyprpaper', 'none']:
            self.options_box.pack_end(self.color_picker_button, False, False, 0)

    def check_backends(self) -> None:
        """Before running the app, check which backends are installed or show the error"""
        if len(self.cf.installed_backends) == 1:
            self.show_message(self.txt.err_backend)
            exit()

    def show_message(self, message: str) -> None:
        """Show messages to user with ok button"""
        dialog = Gtk.MessageDialog(
            parent=self,
            flags=0,
            type=Gtk.MessageType.ERROR,
            buttons=Gtk.ButtonsType.OK,
            message_format=message,
        )
        dialog.run()
        dialog.destroy()


    def process_images(self) -> None:
        """Load images from the selected folder, resize them, and arrange into a grid"""

        self.image_paths = get_image_paths(self.cf.backend, self.cf.image_folder_list, self.cf.include_subfolders,
                                      self.cf.include_all_subfolders, self.cf.show_hidden, self.cf.show_gifs_only)

        # Sort paths:
        if self.cf.sort_option in ["name", "namerev"]:
            self.image_paths.sort(reverse=(self.cf.sort_option == "namerev"))
        if self.cf.sort_option in ["date", "daterev"]:
            self.image_paths.sort(key=lambda x: os.path.getmtime(x), reverse=(self.cf.sort_option == "daterev"))

        # Show caching label:
        self.loading_label = Gtk.Label(label=self.txt.msg_caching)
        self.bottom_loading_box.add(self.loading_label)
        self.bottom_loading_box.show_all()

        self.thumbnails = []
        self.image_names = []

        for image_path in self.image_paths:

            # Skip zero byte files inside the image_path:
            if os.path.getsize(image_path) == 0:
                self.image_paths.remove(image_path)
                continue

            # If this image is not cached yet, resize and cache it:
            cached_image_path = self.cf.cache_dir / os.path.basename(image_path)
            if not cached_image_path.exists():
                cache_image(image_path, self.cf.cache_dir)

            # Load cached thumbnail:
            thumbnail = GdkPixbuf.Pixbuf.new_from_file(str(cached_image_path))
            self.thumbnails.append(thumbnail)

            # Get image name, which may or may not include parent folders:
            image_name = get_image_name(image_path, self.cf.image_folder_list, self.cf.show_path_in_tooltip)
            self.image_names.append(image_name)

        # When image processing is done, remove caching label and display the images:
        self.bottom_loading_box.remove(self.loading_label)
        GLib.idle_add(self.load_image_grid)


    def get_filtered_images(self) -> list:
        """Filter image paths, names, and thumbnails based on the search query, if any"""
        # Read what is written in the search bar, and if nothing, return all images:
        search_query = self.search_entry.get_text().lower()
        if not search_query:
            return self.thumbnails, self.image_names, self.image_paths

        # Otherwise, filter only images whos names match the search query:
        image_names = [name for name in self.image_names if search_query in name.lower()]
        thumbnails = [self.thumbnails[i] for i in range(len(self.image_names)) if search_query in self.image_names[i].lower()]
        image_paths = [self.image_paths[i] for i in range(len(self.image_names)) if search_query in self.image_names[i].lower()]

        return thumbnails, image_names, image_paths


    def load_image_grid(self) -> None:
        """Reload the grid of images"""

        thumbnails, image_names, image_paths = self.get_filtered_images()

        # Clear existing images:
        for child in self.grid.get_children():
            self.grid.remove(child)

        current_y = 0
        current_row_heights = [0] * self.cf.number_of_columns
        for index, [thumbnail, name, path] in enumerate(zip(thumbnails, image_names, image_paths)):

            row = index // self.cf.number_of_columns
            column = index % self.cf.number_of_columns

            # Calculate current y coordinate in the scroll window:
            aspect_ratio = thumbnail.get_width() / thumbnail.get_height()
            current_row_heights[column] = int(240 / aspect_ratio)
            if column == 0:
                current_y += max(current_row_heights) + 10
                current_row_heights = [0]*self.cf.number_of_columns

            # Create a button with an image and add tooltip:
            image = Gtk.Image.new_from_pixbuf(thumbnail)
            image.set_tooltip_text(name)
            button = Gtk.Button()
            if index == self.selected_index:
                button.set_relief(Gtk.ReliefStyle.NORMAL)
                button.get_style_context().add_class("highlighted-button")
                self.highlighted_image_y = current_y
            else:
                button.set_relief(Gtk.ReliefStyle.NONE)
            button.add(image)

            # Add button to the grid and connect clicked event:
            self.grid.attach(button, column, row, 1, 1)
            button.connect("clicked", self.on_image_clicked, path)

        self.grid.show_all()
        self.toggle_zen_mode()


    def toggle_zen_mode(self):
        """Hide or show UI elements when zen mode is enabled or disabled"""
        if self.cf.zen_mode:
            for widget in self.top_container:
                widget.hide()
            for widget in self.options_box:
                widget.hide()
        else:
            for widget in self.top_container:
                widget.show()
            for widget in self.options_box:
                widget.show()


    # def on_window_resize(self, widget, allocation) -> None:
        # """Recalculate the number of columns on window resize and repopulate the grid"""

        # As frequent resize freezed the interface, so we only do it each fifth resize:
        # self.number_of_resize += 1
        # if self.number_of_resize < 5:
            # return

        # Calculate new number of columns and reload the grid:
        # self.cf.number_of_columns = max(1, allocation.width // 250)
        # GLib.idle_add(self.load_image_grid)
        # self.number_of_resize = 0


    def scroll_to_selected_image(self) -> None:
        """Scroll the window to see the highlighted image"""
        scrolled_window_height = self.scrolled_window.get_vadjustment().get_page_size()
        # current_y = self.highlighted_image_row * 180
        subscreen_num = self.highlighted_image_y // scrolled_window_height
        scroll = scrolled_window_height * subscreen_num
        self.scrolled_window.get_vadjustment().set_value(scroll)


    def set_selected_wallpaper(self, path: str) -> None:
        """Set selected image as a wallpaper and save the state"""
        self.swww_options_read()
        self.cf.select_wallpaper(path)
        if self.cf.selected_wallpaper:
            threading.Thread(target=change_wallpaper, args=(self.cf.selected_wallpaper, self.cf, self.cf.selected_monitor)).start()
        self.cf.attribute_selected_wallpaper()
        self.cf.save()


    def choose_folder(self) -> None:
        """Choosing the folder of images, saving the path, and reloading images"""
        dialog = Gtk.FileChooserDialog(
            self.txt.msg_choosefolder, self, Gtk.FileChooserAction.SELECT_FOLDER,
            (Gtk.STOCK_CANCEL, Gtk.ResponseType.CANCEL, self.txt.msg_select, Gtk.ResponseType.OK)
        )
        response = dialog.run()
        if response == Gtk.ResponseType.OK:
            self.cf.image_folder_list = [Path(dialog.get_filename())]
            threading.Thread(target=self.process_images).start()
        dialog.destroy()


    def on_choose_folder_clicked(self, widget) -> None:
        """Choosing the folder of images, saving the path, and reloading images"""
        self.choose_folder()
        self.cf.save()


    def on_filter_gifs_toggled(self, toggle) -> None:
        """Toggle only gifs checkbox via menu"""
        self.cf.show_gifs_only = toggle.get_active()
        threading.Thread(target=self.process_images).start()


    def on_zen_mode_toggled(self, toggle) -> None:
        """Toggle zen mode checkbox via menu"""
        self.show_message("You are entering Zen mode.\nPress z to return to normal mode.")
        self.cf.zen_mode = toggle.get_active()
        self.load_image_grid()


    def on_mpv_sound_toggled(self, toggle) -> None:
        """Toggle sound of mpv player"""
        self.cf.mpvpaper_sound = toggle.get_active()
        subprocess.Popen(f"echo 'cycle mute' | socat - /tmp/mpv-socket-{self.cf.selected_monitor}", shell=True)


    def on_include_subfolders_toggled(self, toggle) -> None:
        """Toggle subfolders visibility via menu"""
        self.cf.include_subfolders = toggle.get_active()
        if not self.cf.include_subfolders:
            self.cf.include_all_subfolders = False
        threading.Thread(target=self.process_images).start()


    def on_include_all_subfolders_toggled(self, toggle) -> None:
        """Toggle subfolders visibility via menu"""
        self.cf.include_all_subfolders = toggle.get_active()
        if self.cf.include_all_subfolders:
            self.cf.include_subfolders = True
        threading.Thread(target=self.process_images).start()


    def toggle_include_subfolders(self) -> None:
        """Toggle subfolders visibility via key"""
        self.cf.include_subfolders = not self.cf.include_subfolders
        threading.Thread(target=self.process_images).start()


    def on_hidden_files_toggled(self, toggle) -> None:
        """Toggle visibility of hidden files via menu"""
        self.cf.show_hidden = toggle.get_active()
        threading.Thread(target=self.process_images).start()


    def toggle_hidden_files(self) -> None:
        """Toggle visibility of hidden files via keys"""
        self.cf.show_hidden = not self.cf.show_hidden
        threading.Thread(target=self.process_images).start()


    def on_show_path_in_tooltip_toggled(self, widget) -> None:
        """Toggle show image relative path in image tooltip"""
        self.cf.show_path_in_tooltip = not self.cf.show_path_in_tooltip
        threading.Thread(target=self.process_images).start()


    def on_fill_option_changed(self, combo) -> None:
        """Save fill parameter when it was changed"""
        self.cf.fill_option = combo.get_active_text().lower()


    def on_monitor_option_changed(self, combo) -> None:
        """Save monitor parameter when it was changed"""
        self.cf.selected_monitor = combo.get_active_text()


    def on_sort_option_changed(self, combo) -> None:
        """Save sort parameter when it is changed"""
        selected_option = combo.get_active_text()
        selected_option_num = list(SORT_DISPLAYS.values()).index(selected_option)
        self.cf.sort_option =  list(SORT_DISPLAYS.keys())[selected_option_num]
        threading.Thread(target=self.process_images).start()


    def on_backend_option_changed(self, combo) -> None:
        """Save backend parameter whet it is changed"""
        self.cf.backend = self.backend_option_combo.get_active_text()
        self.cf.selected_monitor = "All"
        self.monitor_option_display()
        self.mpv_options_display()
        self.fill_option_display()
        self.color_picker_display()
        self.swww_options_display()
        self.show_all()


    def on_transition_option_changed(self, combo) -> None:
        """Update the active transition type based on the selected option"""
        active_index = combo.get_active()
        self.cf.swww_transition_type = SWWW_TRANSITION_TYPES[active_index]
        print(f"Transition type changed to: {self.cf.swww_transition_type}")


    def on_color_set(self, color_button):
        """Convert selected color to web format"""
        rgba_color = color_button.get_rgba()
        red = int(rgba_color.red * 255)
        green = int(rgba_color.green * 255)
        blue = int(rgba_color.blue * 255)
        self.cf.color = "#{:02X}{:02X}{:02X}".format(red, green, blue)


    def on_image_clicked(self, widget, path: str) -> None:
        """On clicking an image, set it as a wallpaper and save"""
        self.selected_index = self.image_paths.index(path)
        self.load_image_grid()
        self.set_selected_wallpaper(path)

    def on_refresh_clicked(self, widget) -> None:
        """On clicking refresh button, clear cache"""
        self.clear_cache()

    def on_mpv_stop_button_clicked(self, widget) -> None:
        """On clicking mpv stop button, kill the mpvpaper"""
        subprocess.Popen(["killall", "mpvpaper"])

    def on_mpv_pause_button_clicked(self, widget) -> None:
        """On clicking mpv stop button, kill the mpvpaper"""
        subprocess.Popen(f"echo 'cycle pause' | socat - /tmp/mpv-socket-{self.cf.selected_monitor}", shell=True)

    def on_random_clicked(self, widget) -> None:
        """On clicking random button, set random wallpaper"""
        self.set_random_wallpaper()

    def on_exit_clicked(self, widget) -> None:
        """On clicking exit button, exit"""
        Gtk.main_quit()


    def set_random_wallpaper(self) -> None:
        """Choose a random image and set it as the wallpaper"""
        new_wallpaper =  get_random_file(self.cf.backend, self.cf.image_folder_list, self.cf.include_subfolders,
                                         self.cf.include_all_subfolders, self.cf.cache_dir)
        if new_wallpaper:
            self.cf.select_wallpaper(new_wallpaper)
        else:
            return
        if self.cf.selected_wallpaper:
            threading.Thread(target=change_wallpaper, args=(self.cf.selected_wallpaper, self.cf, self.cf.selected_monitor)).start()
        self.cf.attribute_selected_wallpaper()
        self.cf.save()


    def clear_cache(self) -> None:
        """Delete cache folder and reprocess the images"""
        try:
            shutil.rmtree(self.cf.cache_dir)
            os.makedirs(self.cf.cache_dir)
        except OSError as e:
            print(f"{self.txt.err_cache} '{self.cf.cache_dir}': {e}")
        threading.Thread(target=self.process_images).start()


    def on_key_pressed(self, widget, event) -> bool:
        """Process various key binding"""

        # Processing keys for losing focus on text fields:
        if self.is_enering_text:
            if event.keyval in [Gdk.KEY_Escape, Gdk.KEY_Return, Gdk.KEY_KP_Enter]:
                self.reset_input_fields()
            return

        # Processing rest of the keys:
        elif (event.keyval == Gdk.KEY_q) or (event.keyval == Gdk.KEY_Escape):
            Gtk.main_quit()

        elif event.keyval == Gdk.KEY_r:
            self.clear_cache()

        elif event.keyval == Gdk.KEY_R:
            self.set_random_wallpaper()

        elif event.keyval in [Gdk.KEY_period]:
            self.toggle_hidden_files()

        elif event.keyval in [Gdk.KEY_slash]:
            self.search_entry.grab_focus()
            return True

        elif event.keyval in [Gdk.KEY_s]:
            self.toggle_include_subfolders()

        elif event.keyval in [Gdk.KEY_h, Gdk.KEY_Left]:
            self.selected_index = max(self.selected_index - 1, 0)
            self.load_image_grid()
            self.scroll_to_selected_image()

        elif event.keyval in [Gdk.KEY_j, Gdk.KEY_Down]:
            self.selected_index = min(self.selected_index + self.cf.number_of_columns, len(self.image_paths) - 1)
            self.load_image_grid()
            self.scroll_to_selected_image()

        elif event.keyval in [Gdk.KEY_k, Gdk.KEY_Up]:
            self.selected_index = max(self.selected_index - self.cf.number_of_columns, 0)
            self.load_image_grid()
            self.scroll_to_selected_image()

        elif event.keyval in [Gdk.KEY_l, Gdk.KEY_Right]:
            self.selected_index = min(self.selected_index + 1, len(self.image_paths) - 1)
            self.load_image_grid()
            self.scroll_to_selected_image()

        elif event.keyval == Gdk.KEY_f:
            self.choose_folder()

        elif event.keyval == Gdk.KEY_g:
            self.selected_index = 0
            self.load_image_grid()
            self.scroll_to_selected_image()

        elif event.keyval == Gdk.KEY_z:
            self.cf.zen_mode = not self.cf.zen_mode
            self.load_image_grid()

        elif event.keyval == Gdk.KEY_G:
            self.selected_index = len(self.image_paths) - 1
            self.load_image_grid()
            self.scroll_to_selected_image()

        elif event.keyval == Gdk.KEY_question:
            message = self.txt.msg_help
            self.show_message(message)

        elif event.keyval == Gdk.KEY_Return or event.keyval == Gdk.KEY_KP_Enter:
            wallpaper_path = self.image_paths[self.selected_index]
            self.set_selected_wallpaper(wallpaper_path)

        # Prevent other default key handling:
        return event.keyval in [Gdk.KEY_Up, Gdk.KEY_Down, Gdk.KEY_Left, Gdk.KEY_Right, Gdk.KEY_Return, Gdk.KEY_KP_Enter, Gdk.KEY_period]


    def on_search_entry_changed(self, entry, event=None):
        """This function is triggered when the user types in the search field"""
        self.load_image_grid()

    def on_clear_button(self, event):
        self.search_entry.set_text("")
        self.main_box.grab_focus()

    def on_focus_in(self, widget, event):
        self.is_enering_text = True

    def on_focus_out(self, widget, event):
        self.is_enering_text = False

    def reset_input_fields(self) -> None:
        """Reset all input fields and remove focus from them"""
        self.search_entry.set_visible(False)
        self.search_entry.set_visible(True)
        self.swww_angle_entry.set_visible(False)
        self.swww_angle_entry.set_visible(True)
        self.swww_steps_entry.set_visible(False)
        self.swww_steps_entry.set_visible(True)
        self.swww_duration_entry.set_visible(False)
        self.swww_duration_entry.set_visible(True)
        self.swww_fps_entry.set_visible(False)
        self.swww_fps_entry.set_visible(True)
        self.main_box.grab_focus()
        self.is_enering_text = False

    def on_window_clicked(self, widget, event) -> bool:
        """Handle clicks outside of input fields to unfocus them"""
        if self.is_enering_text:
            x, y = event.get_coords()

            focused_widget = self.get_focus()
            if focused_widget is not None:
                alloc = focused_widget.get_allocation()
                widget_x, widget_y = focused_widget.translate_coordinates(self, 0, 0) if focused_widget.translate_coordinates(self, 0, 0) else (0, 0)

                if (x < widget_x or x > widget_x + alloc.width or
                    y < widget_y or y > widget_y + alloc.height):
                    self.reset_input_fields()
                    return True

        return False

    def run(self) -> None:
        """Run GUI application"""
        self.connect("destroy", self.on_exit_clicked)
        self.show_all()
        Gtk.main()



"""
This is a daemon that randomly changes wallpaper every specified number of minutes.
THIS IS WORK IN PROGRESS AND NOT USED IN THE PROGRAM YET.
"""

import time
import os
import argparse

def main():
    parser = argparse.ArgumentParser(description="Randomly changes wallpaper every specified number of seconds.")
    parser.add_argument("interval", type=int, help="Time interval in seconds until next wallpaper change.")
    args = parser.parse_args()

    try:
        while True:
            os.system("waypaper --random")
            print(f"Command to change wallpaper executed. Waiting {args.interval} seconds.")
            time.sleep(args.interval)
    except KeyboardInterrupt:
        print("Program interrupted. Exiting.")

if __name__ == "__main__":
    main()



"""Module that runs the system processes to change the wallpaper"""

import subprocess
import time
import screeninfo
from typing import Optional
from pathlib import Path

from waypaper.config import Config


def find_process_pid(command: str) -> Optional[int]:
    """Find the PID of the process matching the exact command"""
    try:
        result = subprocess.run(['ps', 'aux'], stdout=subprocess.PIPE, text=True)
        processes = result.stdout.splitlines()
        for process in processes:
            if command in process:
                # Extract PID (second column after splitting):
                return int(process.split()[1])
        return None
    except Exception:
        return None


def seek_and_destroy(process: str, monitor: str = "All"):
    """Find if a backend is already running somewhere and kill it"""

    # Kill all process instances if we want to set for all monitors:
    if monitor == "All":
        try:
            subprocess.check_output(["pgrep", f"{process}"], encoding='utf-8')
            subprocess.Popen(["killall", f"{process}"])
            time.sleep(0.1)
            print(f"Killed all previous instances of {process}")
        except subprocess.CalledProcessError:
            pass

    # Otherwise, find PID of the process for certain monitor and kill it:
    else:
        if process == "mpvpaper":
            pid = find_process_pid(f"mpvpaper -f socket-{monitor}")
        elif process == "swaybg":
            pid = find_process_pid(f"swaybg -o {monitor}")
        else:
            return
        try:
            subprocess.run(['kill', '-9', str(pid)], check=True)
            print(f"Detected {process} on {monitor} and killed it")
        except Exception as e:
            pass


def change_with_swaybg(image_path: Path, cf: Config, monitor: str):
    """Change wallpaper with swaybg backend"""

    # Check pid of current swaybg process:
    if monitor == "All":
        pid = find_process_pid(f"swaybg")
    else:
        pid = find_process_pid(f"swaybg -o {monitor}")

    # Launch a new swaybg process:
    fill = cf.fill_option.lower()
    command = ["swaybg"]
    if monitor != "All":
        command.extend(["-o", monitor])
    command.extend(["-i", str(image_path)])
    command.extend(["-m", fill, "-c", cf.color])
    subprocess.Popen(command)

    # Kill previous swaybg process once new wallpaper is set:
    if pid:
        time.sleep(0.2)
        subprocess.run(['kill', '-9', str(pid)], check=True)


def change_with_mpvpaper(image_path: Path, cf: Config, monitor: str):
    """Change wallpaper with mpvpaper backend"""

    fill_types = {
            "fill": "panscan=1.0",
            "fit": "panscan=0.0",
            "center": "",
            "stretch": "--keepaspect=no",
            "tile": "",
            }
    fill = fill_types[cf.fill_option.lower()]

    # If mpvpaper is already active on given monitor, try to call that process in that socket:
    try:
        subprocess.check_output(["pgrep", "-f", f"socket-{monitor}"], encoding='utf-8')
        time.sleep(0.2)
        print(f"Detected running mpvpaper on {monitor}, now trying to call mpvpaper socket")
        subprocess.Popen(f"echo 'loadfile \"{image_path}\"' | socat - /tmp/mpv-socket-{monitor}", shell=True)

    # If mpvpaper is not running, create a new process in a new socket:
    except subprocess.CalledProcessError:
        print("Detected no running mpvpaper, starting new mpvpaper process")
        command = ["mpvpaper", "--fork"]
        if cf.mpvpaper_sound:
            command.extend(["-o", f"input-ipc-server=/tmp/mpv-socket-{monitor} {cf.mpvpaper_options} loop {fill} --background-color='{cf.color}'"])
        else:
            command.extend(["-o", f"input-ipc-server=/tmp/mpv-socket-{monitor} {cf.mpvpaper_options} loop {fill} --mute=yes --background-color='{cf.color}'"])

        # Specify the monitor:
        if monitor == "All":
            command.extend('*')
        else:
            command.extend([monitor])

        command.extend([image_path])

        print(f"{command=}")
        subprocess.Popen(command)


def change_with_swww(image_path: Path, cf: Config, monitor: str):
    """Change wallpaper with swww backend"""

    # Because swaybg and hyprpaper are known to conflict with swww, kill them:
    seek_and_destroy("swaybg")
    seek_and_destroy("hyprpaper")

    fill_types = {
            "fill": "crop",
            "fit": "fit",
            "center": "no",
            "stretch": "crop",
            "tile": "no",
            }
    fill = fill_types[cf.fill_option.lower()]

    # Check if swww-daemon is already running. If not, launch it:
    try:
        subprocess.check_output(["pgrep", "swww-daemon"], encoding='utf-8')
    except subprocess.CalledProcessError:
        subprocess.Popen(["swww-daemon"])
        print("Launched swww-daemon")

    command = ["swww", "img", image_path]
    command.extend(["--resize", fill])
    command.extend(["--fill-color", cf.color])
    command.extend(["--transition-type", cf.swww_transition_type])
    command.extend(["--transition-step", str(cf.swww_transition_step)])
    command.extend(["--transition-angle", str(cf.swww_transition_angle)])
    command.extend(["--transition-duration", str(cf.swww_transition_duration)])
    command.extend(["--transition-fps", str(cf.swww_transition_fps)])
    if monitor != "All":
        command.extend(["--outputs", monitor])
    subprocess.Popen(command)


def change_with_feh(image_path: Path, cf: Config, monitor: str):
    """Change wallpaper with feh backend"""

    fill_types = {
            "fill": "--bg-fill",
            "fit": "--bg-max",
            "center": "--bg-center",
            "stretch": "--bg-scale",
            "tile": "--bg-tile",
            }
    fill = fill_types[cf.fill_option.lower()]
    command = ["feh", fill, "--image-bg", cf.color]
    command.extend([str(image_path)])
    subprocess.Popen(command)


def change_with_wallutils(image_path: Path, cf: Config, monitor: str):
    """Change wallpaper with wallutils backend"""
    fill_types = {
            "fill": "scale",
            "fit": "scale",
            "center": "center",
            "stretch": "stretch",
            "tile": "tile",
            }
    fill = fill_types[cf.fill_option.lower()]
    subprocess.Popen(["setwallpaper", "--mode", fill, image_path])


def change_with_hyprpaper(image_path: Path, cf: Config, monitor: str):
    """Change wallpaper with hyprpaper backend"""

    # Check if hyprpaper is already running, otherwise start it, and preload the wallpaper:
    try:
        subprocess.check_output(["pgrep", "hyprpaper"], encoding='utf-8')
    except subprocess.CalledProcessError:
        subprocess.Popen(["hyprpaper"])
        time.sleep(1)
    preload_command = ["hyprctl", "hyprpaper", "preload", image_path]

    # Decide which monitors are affected:
    if monitor == "All":
        monitors = [m.name for m in screeninfo.get_monitors()]
    else:
        monitors: list = [monitor]

    # Change the wallpaper one by one for each affected monitor:
    for m in monitors:
        wallpaper_command = ["hyprctl", "hyprpaper", "wallpaper", f"{m},{image_path}"]
        unload_command = ["hyprctl", "hyprpaper", "unload", "all"]
        result: str = ""
        retry_counter: int = 0

        # Since sometimes hyprpaper fails to change the wallpaper, we try until success:
        while result != "ok" and retry_counter < 10:
            try:
                subprocess.check_output(unload_command, encoding="utf-8").strip()
                subprocess.check_output(preload_command, encoding="utf-8").strip()
                result = subprocess.check_output(wallpaper_command, encoding="utf-8").strip()
                time.sleep(0.1)
            except Exception:
                retry_counter += 1


def change_wallpaper(image_path: Path, cf: Config, monitor: str):
    """Run system commands to change the wallpaper depending on the backend"""

    print(f"Selected file: {image_path}")

    try:
        if cf.backend == "swaybg":
            change_with_swaybg(image_path, cf, monitor)
        if cf.backend == "mpvpaper":
            change_with_mpvpaper(image_path, cf, monitor)
        if cf.backend == "swww":
            change_with_swww(image_path, cf, monitor)
        if cf.backend == "feh":
            change_with_feh(image_path, cf, monitor)
        if cf.backend == "wallutils":
            change_with_wallutils(image_path, cf, monitor)
        if cf.backend == "hyprpaper":
            change_with_hyprpaper(image_path, cf, monitor)
        if cf.backend != "none":
            filename = Path(image_path).resolve().name
            print(f"Sent {cf.backend} command to set {filename} on {monitor} display\n")

        # Run a post command:
        if cf.post_command and cf.use_post_command:
            modified_image_path = str(image_path).replace(" ", "\\ ")
            post_command = cf.post_command.replace("$wallpaper", modified_image_path)
            subprocess.Popen(post_command, shell=True)
            print(f'Executed "{post_command}" post-command\n')

    except Exception as e:
        print(f"Error occured while changing wallpaper: \n{e}")



"""
Module with translations into various languages.
To add a new language, add a new class, and update the load_language() function.
"""


class English:
    def __init__(self):
        self.msg_desc = "GUI wallpaper setter for Wayland and X11. It works as a frontend for feh, swaybg, wallutils, hyprpaper, mpvpaper, and swww."
        self.msg_info = "For more information, visit:\nhttps://github.com/anufrievroman/waypaper"

        self.msg_arg_help = "print version of the program"
        self.msg_arg_fill = "specify how to fill the screen with chosen image"
        self.msg_arg_rest = "restore last wallpaper"
        self.msg_arg_back = "specify which backend to use to set wallpaper"
        self.msg_arg_rand = "set a random wallpaper"
        self.msg_arg_list = "lists some parameters in json to standard out"
        self.msg_arg_wall = "set the specified wallpaper"
        self.msg_arg_folder = "specify which folder to scan for wallpapers"
        self.msg_arg_statefile = "specify a custom file to store the application state"
        self.msg_arg_monitor = "specify desired monitor using its name"
        self.msg_arg_post = "prevents running post_command set in config"
        self.msg_arg_show_path_in_tooltip = "show the relative path in the tooltip"

        self.msg_select = "Select"
        self.msg_refresh = "Refresh"
        self.msg_clear = "Clear"
        self.msg_start = "Start auto-change"
        self.msg_stop = "Stop all"
        self.msg_sound = "Sound"
        self.msg_pause = "Pause"
        self.msg_search = "Search"
        self.msg_random = "Random"
        self.msg_exit = "Exit"
        self.msg_options = "Options"
        self.msg_subfolders = "Show subfolders"
        self.msg_all_subfolders = "Show all subfolders"
        self.msg_hidden = "Show hidden"
        self.msg_gifs = "Show gifs only"
        self.msg_transitions = "Show transition options"
        self.msg_changefolder = "Folder"
        self.msg_choosefolder = "Please choose a folder"
        self.msg_caching = "Caching wallpapers..."
        self.msg_show_path_in_tooltip = "Show path in tooltip"
        self.msg_zen = "Zen mode"
        self.msg_zen_enter = "You are entering Zen mode.\nPress z to return to normal mode."

        self.msg_help = "Waypaper's hotkeys:\n\nhjkl - Navigation (←↓↑→)\nEnter - Set selected wallpaper\nf - Change wallpaper folder\n"
        self.msg_help += "g - Scroll to top\nG - Scroll to bottom\nR - Set random wallpaper\nr - Recache wallpapers\n"
        self.msg_help += "z - Toggle zen mode\n"
        self.msg_help += ". - Toggle hidden images\ns - Toggle images in subfolders\n/ - Search\n? - Help\nq - Exit\n\n"
        self.msg_help += self.msg_info

        self.err_cache = "Error deleting cache"
        self.err_backend = "Looks like none of the wallpaper backends is installed in the system.\n"
        self.err_backend += "Use your package manager to install at least one of these backends:\n\n"
        self.err_backend += "- swaybg (Wayland)\n- swww (Wayland)\n"
        self.err_backend += "- hyprpaper (Wayland)\n- feh (Xorg)\n"
        self.err_backend += "- wallutils (Xorg & Wayland)\n- mpvpaper (Xorg & Wayland)\n\n"
        self.err_backend += self.msg_info

        self.tip_refresh = "Recache the folder of images"
        self.tip_fill = "Choose fill type"
        self.tip_backend = "Choose backend"
        self.tip_sorting = "Choose sorting type"
        self.tip_display = "Choose display"
        self.tip_color = "Choose background color"
        self.tip_random = "Set random wallpaper"
        self.tip_exit = "Exit the application"
        self.tip_transition = "Choose transition type"
        self.tip_mpv_stop = "Stop all all mpv processes"
        self.tip_mpv_pause = "Play/Pause video wallpaper"
        self.tip_mpv_sound = "Play sound of the video"
        self.tip_timer = "How often to automatically change wallpaper"
        self.tip_start = "Initiate automatic change of wallpaper after a time interval"


class German:
    def __init__(self):
        self.msg_desc = "Grafisches Hintergrundbild-Auswahlwerkzeug für Wayland und X11. Es dient als Frontend für feh, swaybg, wallutils, hyprpaper, mpvpaper, und swww."
        self.msg_info = "Weitere Informationen finden Sie unter:\nhttps://github.com/anufrievroman/waypaper"

        self.msg_arg_help = "gibt die Programmversion aus"
        self.msg_arg_fill = "legt fest, wie das Bild skaliert werden soll, um den gesamten Bildschirm auszufüllen"
        self.msg_arg_rest = "stellt das zuletzt verwendete Hintergrundbild wieder her"
        self.msg_arg_back = "legt das Backend fest, welches zum Setzen des Hintergrundbildes verwendet werden soll"
        self.msg_arg_rand = "wählt ein zufälliges Hintergrundbild aus"
        self.msg_arg_list = "lists some parameters in json to standard out"
        self.msg_arg_wall = "setzt das angegebene Hintergrundbild"
        self.msg_arg_folder = "legt fest, welcher Ordner nach Hintergrundbildern durchsucht werden soll"
        self.msg_arg_statefile = "specify a custom file to store the application state"
        self.msg_arg_monitor = "geben Sie den gewünschten Monitor mit seinem Namen an"
        self.msg_arg_post = "verhindert das Ausführen des in der Konfiguration festgelegten post_command"
        self.msg_arg_show_path_in_tooltip = "zeigt den relativen Pfad im Tooltip an"
        self.msg_show_path_in_tooltip = "Pfad im Tooltip anzeigen"

        self.msg_select = "Auswählen"
        self.msg_refresh = "Aktualisieren"
        self.msg_random = "Zufällig"
        self.msg_clear = "Löschen"
        self.msg_start = "Start auto-change"
        self.msg_stop = "Stop all"
        self.msg_sound = "Sound"
        self.msg_pause = "Pause"
        self.msg_search = "Suchen"
        self.msg_options = "Optionen"
        self.msg_exit = "Beenden"
        self.msg_subfolders = "Unterordner"
        self.msg_all_subfolders = "Show all subfolders"
        self.msg_hidden = "Zeige versteckte"
        self.msg_gifs = "Nur GIFs anzeigen"
        self.msg_transitions = "Übergangsoptionen anzeigen"
        self.msg_changefolder = "Ordner"
        self.msg_choosefolder = "Bitte wählen Sie einen Ordner aus"
        self.msg_caching = "Hintergrundbilder werden zwischengespeichert..."
        self.msg_show_path_in_tooltip = "Pfad in Tooltip anzeigen"
        self.msg_zen = "Zen-Modus"
        self.msg_zen_enter = "Du betrittst den Zen-Modus.\nDrücke z, um in den normalen Modus zurückzukehren."

        self.msg_help = "Waypapers Tastenkürzel:\n\nhjkl - Navigation (←↓↑→)\nf - Hintergrundbild-Ordner ändern\n"
        self.msg_help += "g - Zum Anfang scrollen\nG - Zum Ende scrollen\nR - Zufälliges Hintergrundbild\nr - Hintergrundbilder zwischenspeichern\n"
        self.msg_help += "z - Toggle zen mode\n"
        self.msg_help += ". - Versteckte Bilder einbeziehen/ausschließen\ns - Unterordner mit einbeziehen\n/ - Search\n? - Hilfe\nq - Beenden\n\n"
        self.msg_help += self.msg_info

        self.err_cache = "Fehler beim Löschen des Zwischenspeichers"
        self.err_backend = "Es konnte kein Hintergrundbild-Backend gefunden werden.\n"
        self.err_backend += "Installieren Sie mindestens eines der folgenden Backends:\n\n"
        self.err_backend += "- swaybg (Wayland)\n- swww (Wayland)\n"
        self.err_backend += "- hyprpaper (Wayland)\n- feh (Xorg)\n"
        self.err_backend += "- wallutils (Xorg & Wayland)\n- mpvpaper (Xorg & Wayland)\n\n"
        self.err_backend += self.msg_info

        self.tip_refresh = "Erneutes einlesen des Hintergrundbild-Ordners"
        self.tip_fill = "Skalierungsart auswählen"
        self.tip_backend = "Backend auswählen"
        self.tip_sorting = "Sortierungsart auswählen"
        self.tip_display = "Bildschirm auswählen"
        self.tip_color = "Hintergrundfarbe auswählen"
        self.tip_random = "Ein zufälliges Hintergrundbild auswählen"
        self.tip_exit = "Das Programm beenden"
        self.tip_transition = "Übergangstyp auswählen"
        self.tip_mpv_stop = "Stop allpe alle mpv-Prozesse"
        self.tip_mpv_pause = "Pause Video-Wallpaper"
        self.tip_mpv_sound = "Ton des Videos abspielen"
        self.tip_timer = "Wie oft das Hintergrundbild automatisch gewechselt wird"
        self.tip_start = "Automatischen Hintergrundbildwechsel nach einem Zeitintervall starten"

class French:
    def __init__(self):
        self.msg_desc = "Sélecteur de papier peint graphique pour Wayland et X11. Il fonctionne comme une interface pour feh, swaybg, wallutils, hyprpaper, mpvpaper, et swww."
        self.msg_info = "Pour plus d'informations, visitez :\nhttps://github.com/anufrievroman/waypaper"

        self.msg_arg_help = "afficher la version du programme"
        self.msg_arg_fill = "spécifier comment remplir l'écran avec l'image choisie"
        self.msg_arg_rest = "restaurer le dernier papier peint"
        self.msg_arg_back = "spécifier quel backend utiliser pour définir le papier peint"
        self.msg_arg_rand = "définir un papier peint aléatoire"
        self.msg_arg_list = "liste certains paramètres en JSON vers la sortie standard"
        self.msg_arg_wall = "définir le fond d’écran spécifié"
        self.msg_arg_folder = "spécifier le dossier à analyser pour les fonds d’écran"
        self.msg_arg_statefile = "spécifier un fichier personnalisé pour enregistrer l’état de l’application"
        self.msg_arg_monitor = "spécifiez le moniteur souhaité en utilisant son nom"
        self.msg_arg_post = "empêche l'exécution de post_command défini dans la configuration"
        self.msg_arg_show_path_in_tooltip = "afficher le chemin relatif dans l'infobulle"

        self.msg_select = "Sélectionner"
        self.msg_refresh = "Actualiser"
        self.msg_random = "Aléatoire"
        self.msg_clear = "Effacer"
        self.msg_start = "Start auto-change"
        self.msg_stop = "Arrêter"
        self.msg_sound = "Sound"
        self.msg_pause = "Pause"
        self.msg_search = "Rechercher"
        self.msg_options = "Options"
        self.msg_exit = "Quitter"
        self.msg_subfolders = "Sous-dossiers"
        self.msg_all_subfolders = "Show all subfolders"
        self.msg_hidden = "Afficher caché"
        self.msg_gifs = "Afficher uniquement les gifs"
        self.msg_transitions = "Afficher les options de transition"
        self.msg_changefolder = "Dossier"
        self.msg_choosefolder = "Veuillez choisir un dossier"
        self.msg_caching = "Mise en cache des papiers peints..."
        self.msg_show_path_in_tooltip = "Afficher le chemin dans l'infobulle"
        self.msg_zen = "Mode Zen"
        self.msg_zen_enter = "Vous entrez en mode Zen.\nAppuyez sur z pour revenir au mode normal."

        self.msg_help = "Raccourcis clavier de Waypaper :\n\nhjkl - Navigation (←↓↑→)\nf - Changer de dossier de papier peint\n"
        self.msg_help += "g - Faire défiler vers le haut\nG - Faire défiler vers le bas\nR - Définir un papier peint aléatoire\nr - Recréer le cache des papiers peints\n"
        self.msg_help += "z - Toggle zen mode\n"
        self.msg_help += ". - Inclure/exclure les images cachées\ns - Inclure/exclure les images des sous-dossiers\n/ - Search\n? - Aide\nq - Quitter\n\n"
        self.msg_help += self.msg_info

        self.err_cache = "Erreur lors de la suppression du cache"
        self.err_backend = "Il semble qu'aucun des backends de papier peint ne soit installé sur le système.\n"
        self.err_backend += "Utilisez votre gestionnaire de paquets pour installer au moins l'un de ces backends :\n\n"
        self.err_backend += "- swaybg (Wayland)\n- swww (Wayland)\n"
        self.err_backend += "- hyprpaper (Wayland)\n- feh (Xorg)\n"
        self.err_backend += "- wallutils (Xorg & Wayland)\n- mpvpaper (Xorg & Wayland)\n\n"
        self.err_backend += self.msg_info

        self.tip_refresh = "Recréer le dossier d'images"
        self.tip_fill = "Choisir le type de remplissage"
        self.tip_backend = "Choisir le backend"
        self.tip_sorting = "Choisir le type de tri"
        self.tip_display = "Choisir l'affichage"
        self.tip_color = "Choisir la couleur de fond"
        self.tip_random = "Définir un papier peint aléatoire"
        self.tip_exit = "Quitter l'application"
        self.tip_transition = "Choisissez le type de transition"
        self.tip_mpv_stop = "Arrêter tous les processus mpv"
        self.tip_mpv_pause = "Pause du fond d'écran vidéo"
        self.tip_mpv_sound = "Lire le son de la vidéo"
        self.tip_timer = "Fréquence de changement automatique du fond d’écran"
        self.tip_start = "Lancer le changement automatique du fond d’écran après un intervalle de temps"

class Polish:
    def __init__(self):
        self.msg_desc = "GUI narzędzie do ustawiania tapety dla Wayland i X11. Jest interfejsem dla feh, swaybg, wallutils, hyprpaper, mpvpaper i swww."
        self.msg_info = "Aby dowiedzieć się więcej, odwiedź:\nhttps://github.com/anufrievroman/waypaper"

        self.msg_arg_help = "pokaż wersję programu"
        self.msg_arg_fill = "ustaw jak wypełnić ekran wybranym obrazem"
        self.msg_arg_rest = "przywróć ostatnią tapetę"
        self.msg_arg_back = "ustaw zaplecze do wyświetlania tapety"
        self.msg_arg_rand = "ustaw losową tapetę"
        self.msg_arg_list = "wypisz tapety na standardowe wyjście w formacie json"
        self.msg_arg_wall = "ustaw wybraną tapetę"
        self.msg_arg_folder = "ustaw folder do przeskanowania szukając tapet"
        self.msg_arg_statefile = "ustaw plik, żeby przechowywać stan aplikacji"
        self.msg_arg_monitor = "ustaw pożądany monitor używając jego nazwy"
        self.msg_arg_post = "zapobiega uruchomieniu post_command ustawionego w konfiguracji"
        self.msg_arg_show_path_in_tooltip = "pokaż względną ścieżkę w podpowiedzi"

        self.msg_select = "Wybierz"
        self.msg_refresh = "Odświerz"
        self.msg_clear = "Wyczyść"
        self.msg_start = "Start auto-change"
        self.msg_stop = "Stop all"
        self.msg_sound = "Dźwięk"
        self.msg_pause = "Pauza"
        self.msg_search = "Szukaj"
        self.msg_random = "Losowa"
        self.msg_exit = "Wyjdź"
        self.msg_options = "Opcje"
        self.msg_subfolders = "Pokaż podfolder"
        self.msg_all_subfolders = "Pokaż wszystkie podfoldery"
        self.msg_hidden = "Pokaż ukryte"
        self.msg_gifs = "Pokaż tylko Gify"
        self.msg_transitions = "Pokaż opcje przejścia"
        self.msg_changefolder = "Folder"
        self.msg_choosefolder = "Proszę wybrać folder"
        self.msg_caching = "Buforowanie tapet..."
        self.msg_show_path_in_tooltip = "Pokaż ścieżkę w podpowiedzi"
        self.msg_zen = "Tryb Zen"
        self.msg_zen_enter = "Wchodzisz w tryb Zen.\nNaciśnij z, aby wrócić do trybu normalnego."

        self.msg_help = "Skróty klawiszowe Waypaper'a:\n\nhjkl - Nawigacja (←↓↑→)\nEnter - Ustaw wybraną tapetę\nf - Zmień folder tapet\n"
        self.msg_help += "g - Przewiń na górę\nG - Przewiń na dół\nR - Ustaw losową tapetę\nr - Jeszcze raz buferuj tapety\n"
        self.msg_help += "z - Toggle zen mode\n"
        self.msg_help += ". - Pokaż/ukryj ukryte tapety\ns - Włącz/wyłącz tapety w podfolderach\n/ - Szukanie\n? - Pomoc\nq - Wyjście\n\n"
        self.msg_help += self.msg_info

        self.err_cache = "Błąd w usuwaniu bufera"
        self.err_backend = "Żadne zaplecze nie zostało znalezione.\n"
        self.err_backend += "Use your package manager to install at least one of these backends:\n\n"
        self.err_backend += "Użyj twojego instalera, żeby zainstalować co najmiej jedno zaplecze:\n\n"
        self.err_backend += "- swaybg (Wayland)\n- swww (Wayland)\n"
        self.err_backend += "- hyprpaper (Wayland)\n- feh (Xorg)\n"
        self.err_backend += "- wallutils (Xorg & Wayland)\n- mpvpaper (Xorg & Wayland)\n\n"
        self.err_backend += self.msg_info

        self.tip_refresh = "Ponownie buferuj folder tapet"
        self.tip_fill = "Ustaw sposób wypełnienia"
        self.tip_backend = "Wybierz zaplecze"
        self.tip_sorting = "Wybierz sposób sortowania"
        self.tip_display = "Wybierz monitor"
        self.tip_color = "Ustaw kolor tła"
        self.tip_random = "Ustaw losową tapetę"
        self.tip_exit = "Wyjdź z aplikacji"
        self.tip_transition = "Ustaw sposób przejścia"
        self.tip_mpv_stop = "Zatrzymaj wszystkie procesy mpv"
        self.tip_mpv_pause = "Odtwarzaj/Zatrzymaj animowaną"
        self.tip_mpv_sound = "Odtwórz dźwięk z wideo"
        self.tip_timer = "Jak często automatycznie zmieniać tapetę"
        self.tip_start = "Rozpocznij automatyczną zmianę tapety po upływie określonego czasu"

class Russian:
    def __init__(self):
        self.msg_desc = "Графический интерфейс для установки обоев на Wayland и X11. Работает как фронтенд для feh, swaybg, wallutils, hyprpaper, mpvpaper и swww."
        self.msg_info = "Для получения дополнительной информации посетите:\nhttps://github.com/anufrievroman/waypaper"

        self.msg_arg_help = "вывести версию программы"
        self.msg_arg_fill = "указать, как заполнить экран выбранным изображением"
        self.msg_arg_rest = "восстановить последние обои"
        self.msg_arg_back = "указать бэкенд для установки обоев"
        self.msg_arg_rand = "установить случайные обои"
        self.msg_arg_list = "вывести различные параметры в формате json"
        self.msg_arg_wall = "указать путь к изображению"
        self.msg_arg_folder = "указать путь к папке с обоями"
        self.msg_arg_statefile = "указать путь к файлу состояния"
        self.msg_arg_monitor = "указать имя монитора для которого устанавливаются обои"
        self.msg_arg_post = "предотвратить выполнение post_command"
        self.msg_arg_show_path_in_tooltip = "показывать относительный путь в подсказке"
        self.msg_zen = "Режим Дзэн"
        self.msg_zen_enter = "Вы входите в режим Дзэн.\nНажмите z, чтобы вернуться в обычный режим."

        self.msg_select = "Выбрать"
        self.msg_refresh = "Обновить"
        self.msg_random = "Случайно"
        self.msg_clear = "Очистить"
        self.msg_start = "Начать авто-смену"
        self.msg_stop = "Стоп"
        self.msg_sound = "Звук"
        self.msg_pause = "Пауза"
        self.msg_search = "Поиск"
        self.msg_options = "Опции"
        self.msg_exit = "Выход"
        self.msg_subfolders = "Показать подпапки"
        self.msg_all_subfolders = "Show all subfolders"
        self.msg_hidden = "Показать скрытые"
        self.msg_gifs = "Показать только gif"
        self.msg_transitions = "Показать опции перехода"
        self.msg_changefolder = "Папка"
        self.msg_choosefolder = "Пожалуйста, выберите папку"
        self.msg_caching = "Кэширование обоев..."
        self.msg_show_path_in_tooltip = "Показыват путь в подсказке"

        self.msg_help = "Горячие клавиши Waypaper:\n\nhjkl - Навигация (←↓↑→)\nf - Изменить папку с обоями\n"
        self.msg_help += "g - Прокрутка в начало\nG - Прокрутка в конец\nR - Установить случайные обои\nr - Обновить кэш обоев\n"
        self.msg_help += "z - Переключать zen режим\n"
        self.msg_help += ". - Показать/скрыть скрытые файлы \ns - Показать/скрыть вложенные папки\n/ - Поиск\n? - Справка\nq - Выход\n\n"
        self.msg_help += self.msg_info

        self.err_cache = "Ошибка при удалении кэша"
        self.err_backend = "Похоже, что ни один из бэкендов для установки обоев не установлен в системе.\n"
        self.err_backend += "Используйте менеджер пакетов для установки хотя бы одного из следующих бэкендов:\n\n"
        self.err_backend += "- swaybg (Wayland)\n- swww (Wayland)\n"
        self.err_backend += "- hyprpaper (Wayland)\n- feh (Xorg)\n"
        self.err_backend += "- wallutils (Xorg & Wayland)\n- mpvpaper (Xorg & Wayland)\n\n"
        self.err_backend += self.msg_info

        self.tip_refresh = "Обновить папку с изображениями"
        self.tip_fill = "Выбрать тип заполнения"
        self.tip_backend = "Выбрать бэкенд"
        self.tip_sorting = "Выбрать тип сортировки"
        self.tip_display = "Выбрать дисплей"
        self.tip_color = "Выбрать цвет фона"
        self.tip_random = "Установить случайные обои"
        self.tip_exit = "Выйти из приложения"
        self.tip_transition =  "Выберите тип перехода"
        self.tip_mpv_stop = "Остановить все mpv процессы"
        self.tip_mpv_pause = "Плей/пауза видео-обоев"
        self.tip_mpv_sound = "Проигрывать звук видео-обоев"
        self.tip_timer = "Как часто автоматически менять обои"
        self.tip_start = "Initiate automatic change of wallpaper after a time interval"


class Belarusian:
    def __init__(self):
        self.msg_desc = "Графічны інтэрфейс для ўстаноўкі шпалер на Wayland і X11. Працуе як фронтэнд для feh, swaybg, wallutils, hyprpaper, mpvpaper і swww."
        self.msg_info = "Для атрымання дадатковай інфармацыі наведайце:\nhttps://github.com/anufrievroman/waypaper"

        self.msg_arg_help = "вывесці версію праграмы"
        self.msg_arg_fill = "пазначыць, як запоўніць экран абранай выявай"
        self.msg_arg_rest = "аднавіць апошнія шпалеры"
        self.msg_arg_back = "паказаць бэкенд для ўстаноўкі шпалер"
        self.msg_arg_rand = "ўсталяваць выпадковыя шпалеры"
        self.msg_arg_list = "вывесці шпалеры і матыноры ў фармаце json"
        self.msg_arg_wall = "пазначыць шлях да выявы"
        self.msg_arg_folder = "ўкажыце, які тэчку сканаваць для выявы"
        self.msg_arg_statefile = "ўкажыце карыстацкі файл для захавання стану прыкладання"
        self.msg_arg_monitor = "określ żądany monitor, używając jego nazwy"
        self.msg_arg_post = "перашкаджае запуску post_command, зададзенага ў канфігурацыі"
        self.msg_arg_show_path_in_tooltip = "паказваць адносны шлях у падказцы"

        self.msg_select = "Выбраць"
        self.msg_refresh = "Абнавіць"
        self.msg_random = "Выпадкова"
        self.msg_clear = "Ачысціць"
        self.msg_start = "Start auto-change"
        self.msg_stop = "Стоп"
        self.msg_sound = "Звук"
        self.msg_pause = "Паўза"
        self.msg_search = "Шукаць"
        self.msg_options = "Опцыі"
        self.msg_exit = "Вынахад"
        self.msg_subfolders = "Паказаць падтэчкі"
        self.msg_all_subfolders = "Show all subfolders"
        self.msg_hidden = "Паказаць схаваныя"
        self.msg_gifs = "Паказаць толькі gif"
        self.msg_transitions = "Show transition options"
        self.msg_changefolder = "Тэчка"
        self.msg_choosefolder = "Калі ласка, абярыце тэчку"
        self.msg_caching = "Кэшаванне шпалер..."
        self.msg_show_path_in_tooltip = "Show path in tooltip"
        self.msg_zen = "Рэжым Дзен"
        self.msg_zen_enter = "Вы ўваходзіце ў рэжым Дзен.\nНацісніце z, каб вярнуцца ў звычайны рэжым."

        self.msg_help = "Гарачыя клавішы Waypaper:\n\nhjkl - Навігацыя (←↓↑→)\nf - Змяніць тэчку са шпалерамі\n"
        self.msg_help += "g - Пракрутка ў пачатак\nG - Пракрутка ў канец\nR - Усталяваць выпадковыя шпалеры\nr - Абнавіць кэш шпалер\n"
        self.msg_help += "z - Toggle zen mode\n"
        self.msg_help += ". - Паказаць/схаваць схаваныя файлы \ns - Паказаць/схаваць укладзеныя тэчкі\n/ - Пошук\n? - Даведка\nq - Вынахад\n\n"
        self.msg_help += self.msg_info

        self.err_cache = "Памылка пры выдаленні кэша"
        self.err_backend = "Падобна, што ніводны з бэкэндаў для ўсталёўкі шпалер не ўсталяваны ў сістэме.\n"
        self.err_backend += "Выкарыстоўвайце менеджэр пакетаў для ўстаноўкі хаця б аднаго з наступных бэкендаў:\n\n"
        self.err_backend += "- swaybg (Wayland)\n- swww (Wayland)\n"
        self.err_backend += "- hyprpaper (Wayland)\n- feh (Xorg)\n"
        self.err_backend += "- wallutils (Xorg & Wayland)\n- mpvpaper (Xorg & Wayland)\n\n"
        self.err_backend += self.msg_info

        self.tip_refresh = "Абнавіць тэчку з выявамі"
        self.tip_fill = "Выбраць тып запаўнення"
        self.tip_backend = "Выбраць бэкенд"
        self.tip_sorting = "Выбраць тып сартавання"
        self.tip_display = "Выбраць дысплей"
        self.tip_color = "Выбраць колер фону"
        self.tip_random = "Усталяваць выпадковыя шпалеры"
        self.tip_exit = "Выйсці з прыкладання"
        self.tip_transition = "Выберыце тып пераходу"
        self.tip_mpv_stop = "Спыніць усе працэсы mpv"
        self.tip_mpv_pause = "Паўза відэа-абояў"
        self.tip_mpv_sound = "Прайграваць гук відэа"
        self.tip_timer = "Як часта аўтаматычна змяняць шпалеры"
        self.tip_start = "Пачаць аўтаматычную змену шпалераў пасля пэўнага інтэрвалу часу"


class Chinese:
    def __init__(self):
        self.msg_desc = "Wayland 和 X11 的 GUI 壁纸设置器。它用作 feh、swaybg、hyprpaper、mpvpaper、wallutils 和 swww 的前端。"
        self.msg_info = "欲了解更多信息，请访问:\nhttps://github.com/anufrievroman/waypaper"

        self.msg_arg_help = "版本信息"
        self.msg_arg_fill = "指定所选图像填充屏幕"
        self.msg_arg_rest = "恢复上个壁纸"
        self.msg_arg_back = "指定使用哪个后端来设置壁纸"
        self.msg_arg_rand = "设置随机壁纸"
        self.msg_arg_list = "以 JSON 格式列出壁纸并输出到标准输出"
        self.msg_arg_wall = "设置指定的壁纸"
        self.msg_arg_folder = "指定扫描壁纸的文件夹"
        self.msg_arg_statefile = "指定用于存储应用程序状态的自定义文件"
        self.msg_arg_monitor = "通过其名称指定所需的显示器"
        self.msg_arg_post = "阻止运行配置中设置的 post_command"
        self.msg_arg_show_path_in_tooltip = "在工具提示中显示相对路径"

        self.msg_select = "选择"
        self.msg_refresh = "刷新"
        self.msg_random = "随机"
        self.msg_clear = "清除"
        self.msg_start = "Start auto-change"
        self.msg_stop = "停止"
        self.msg_sound = "Sound"
        self.msg_pause = "暂停"
        self.msg_search = "搜索"
        self.msg_options = "选项"
        self.msg_exit = "退出"
        self.msg_subfolders = "子文件夹"
        self.msg_all_subfolders = "Show all subfolders"
        self.msg_hidden = "显示隐藏项"
        self.msg_gifs = "仅显示GIF"
        self.msg_transitions = "显示过渡选项"
        self.msg_changefolder = "文件夹"
        self.msg_choosefolder = "请选择一个文件夹"
        self.msg_caching = "缓存壁纸..."
        self.msg_show_path_in_tooltip = "在工具提示中显示路径"
        self.msg_zen = "禅模式"
        self.msg_zen_enter = "您正在进入禅模式。\n按 z 返回正常模式。"

        self.msg_help = "Waypaper 的热键：\n\nhjkl -导航 (←↓↑→)\nf -更改壁纸文件夹\n"
        self.msg_help += "g -滚动到顶部\nG -滚动到底部\nR -设置随机壁纸\nr -重新缓存壁纸\n"
        self.msg_help += "z - Toggle zen mode\n"
        self.msg_help += ". - 包括/排除隐藏图像\ns -包含/排除子文件夹中的图像\n/ - Search\n？ -帮助\nq -退出\n\n"
        self.msg_help += self.msg_info

        self.err_cache = "删除缓存时出错"
        self.err_backend = "系统中似乎没有安装壁纸后端。\n"
        self.err_backend += "使用包管理器安装至少以下后端之一：\n\n"
        self.err_backend += "- swaybg (Wayland)\n- swww (Wayland)\n"
        self.err_backend += "- hyprpaper (Wayland)\n- feh (Xorg)\n"
        self.err_backend += "- wallutils (Xorg & Wayland)\n- mpvpaper (Xorg & Wayland)\n\n"
        self.err_backend += self.msg_info

        self.tip_refresh = "重新缓存图像文件夹"
        self.tip_fill = "选择填充类型"
        self.tip_backend = "选择后端"
        self.tip_sorting = "选择排序类型"
        self.tip_display = "选择显示"
        self.tip_color = "选择背景颜色"
        self.tip_random = "设置随机壁纸"
        self.tip_exit = "退出应用程序"
        self.tip_transition = "选择过渡类型"
        self.tip_mpv_stop = "停止所有 mpv 进程"
        self.tip_mpv_pause = "暂停视频壁纸"
        self.tip_mpv_sound = "播放视频的声音"
        self.tip_timer = "自动更换壁纸的频率"
        self.tip_start = "在设定的时间间隔后启动自动更换壁纸"

class TraditionalChinese:
    def __init__(self):
        self.msg_desc = "Wayland 和 X11 的 GUI 壁紙選擇器。是能作為 feh、swaybg、wallutils、hyprpaper、mpvpaper 和 swww 的前端。"
        self.msg_info = "如要更多資訊, 請前往:\nhttps://github.com/anufrievroman/waypaper"

        self.msg_arg_help = "版本"
        self.msg_arg_fill = "壁紙填滿螢幕的方式"
        self.msg_arg_rest = "重回上一個壁紙"
        self.msg_arg_back = "選擇壁紙後端"
        self.msg_arg_rand = "隨機壁紙"
        self.msg_arg_list = "以 JSON 的格式列出壁紙及輸出到標準輪出"
        self.msg_arg_wall = "使用選擇了的壁紙"
        self.msg_arg_folder = "存有壁紙的文件夾"
        self.msg_arg_statefile = "指定用於儲存程式狀態的自定文件"
        self.msg_arg_monitor = "用螢幕的名稱來選擇壁紙在那一個螢幕顯示"
        self.msg_arg_post = "停止更換壁紙後運行的命令"
        self.msg_arg_show_path_in_tooltip = "在工具提示中顯示相對路徑"

        self.msg_select = "選擇"
        self.msg_refresh = "刷新"
        self.msg_clear = "清除搜尋"
        self.msg_start = "開始自動更換"
        self.msg_stop = "停止"
        self.msg_sound = "聲音開關"
        self.msg_pause = "暫停"
        self.msg_search = "搜尋"
        self.msg_random = "隨機"
        self.msg_exit = "退出"
        self.msg_options = "設定"
        self.msg_subfolders = "顯示子文件夾的圖片"
        self.msg_all_subfolders = "顯示所有子文件夾的圖片"
        self.msg_hidden = "顯示隱藏子的圖片"
        self.msg_gifs = "只顯示 GIF 圖"
        self.msg_transitions = "顯示翻譯設定"
        self.msg_changefolder = "文件夾"
        self.msg_choosefolder = "選擇文件夾"
        self.msg_caching = "緩存壁紙"
        self.msg_show_path_in_tooltip = "在工具提示中顯示路徑"
        self.msg_zen = "禪模式"
        self.msg_zen_enter = "您正在進入禪模式。\n按 z 返回正常模式。"

        self.msg_help = "Waypaper 的快捷鍵:\n\nhjkl - 行動 (←↓↑→)\nEnter - 設定選擇壁紙\nf - 選擇壁紙文件夾\n"
        self.msg_help += "g - 滾動到頂部\nG - 滾動到底部\nR - 隨機壁紙\nr - 重新玩全壁紙\n"
        self.msg_help += "z - Toggle zen mode\n"
        self.msg_help += ". - 顯示隱藏的圖片\ns - 顯示在子文件夾的圖片\n/ - Search\n? - 協助\nq - 退出\n\n"
        self.msg_help += self.msg_info

        self.err_cache = "刪除緩存時出錯"
        self.err_backend = "沒有安裝任何壁紙後端.\n"
        self.err_backend += "請使用包裝管理器安裝至少一個後端:\n\n"
        self.err_backend += "- swaybg (Wayland)\n- swww (Wayland)\n"
        self.err_backend += "- hyprpaper (Wayland)\n- feh (Xorg)\n"
        self.err_backend += "- wallutils (Xorg & Wayland)\n- mpvpaper (Xorg & Wayland)\n\n"
        self.err_backend += self.msg_info

        self.tip_refresh = "重新緩存在文件夾中的圖片"
        self.tip_fill = "選擇壁紙填滿方式"
        self.tip_backend = "選擇後端"
        self.tip_sorting = "選擇排序方式"
        self.tip_display = "選擇螢幕"
        self.tip_color = "選擇背景顏色"
        self.tip_random = "隨機壁紙"
        self.tip_exit = "退出程式"
        self.tip_transition = "選擇過渡方式"
        self.tip_mpv_stop = "暫停所有 mpv 程序"
        self.tip_mpv_pause = "開始/暫停影片壁紙播放"
        self.tip_mpv_sound = "播放影片聲音"
        self.tip_timer = "自動選擇壁紙的時間"
        self.tip_start = "開始定時更換壁紙"


class Spanish:
    def __init__(self):
        self.msg_desc = 'Cambiador de fondo de pantalla para "Wayland" y "X11". Trabaja como una interfaz gráfica para "feh", "swaybg", "wallutils", "swww" y "hyprpaper".'
        self.msg_info = "Para más información, visita:\nhttps://github.com/anufrievroman/waypaper"

        self.msg_arg_help = "imprime la versión del programa"
        self.msg_arg_fill = "específica una forma de rellenar la pantalla con la imagen escogida"
        self.msg_arg_rest = "restaura la última imagen de fondo"
        self.msg_arg_back = "específica cuál es el programa a utilizar para cambiar la imagen de fondo"
        self.msg_arg_rand = "aplica una imagen de fondo aleatoria"
        self.msg_arg_list = 'imprime un listado de las imágenes de fondo al terminal en formato "JSON"'
        self.msg_arg_wall = "establece el fondo especificado"
        self.msg_arg_folder = "specify which folder to scan for wallpapers"
        self.msg_arg_statefile = "specify a custom file to store the application state"
        self.msg_arg_monitor = "especifique el monitor deseado usando su nombre"
        self.msg_arg_post = "impide ejecutar el post_command definido en la configuración"
        self.msg_arg_show_path_in_tooltip = "mostrar la ruta relativa en la información sobre herramientas"

        self.msg_select = "Selecciona"
        self.msg_refresh = "Actualizar"
        self.msg_random = "Aleatorio"
        self.msg_clear = "Borrar"
        self.msg_start = "Start auto-change"
        self.msg_stop = "Detener"
        self.msg_sound = "Sonido"
        self.msg_pause = "Pausa"
        self.msg_search = "Buscar"
        self.msg_options = "Opciones"
        self.msg_exit = "Salir"
        self.msg_subfolders = "Ver subcarpetas"
        self.msg_all_subfolders = "Busqueda recursiva"
        self.msg_hidden = "Ver archivos ocultos"
        self.msg_gifs = 'Ver solamente imágenes de tipo "GIF"'
        self.msg_transitions = "Mostrar opciones de transición"
        self.msg_changefolder = "Carpeta"
        self.msg_choosefolder = "Por favor, selecciona una carpeta"
        self.msg_caching = "Almacenando en el caché..."
        self.msg_show_path_in_tooltip = "Mostrar la ruta en la información sobre herramientas"
        self.msg_zen = "Modo Zen"
        self.msg_zen_enter = "Estás entrando en el modo Zen.\nPresiona z para volver al modo normal."

        self.msg_help = 'Controles para usar "Waypaper":\n\nhjkl - Navegación (←↓↑→)\n"Enter" (⏎) - Actualizar imagen de fondo a la imágen seleccionada\nf - Cambiar carpeta de imágenes\n'
        self.msg_help += "g - Ir a la parte de arriba\nG - Ir a la parte de abajo\nR - Cambiar imagen de fondo a una imágen aleatoria\nr - Recrear caché de imágenes\n"
        self.msg_help += "z - Toggle zen mode\n"
        self.msg_help += ". - Ver/Omitir archivos ocultos\ns - Ver/Omitir imágenes en subcarpetas\n/ - Search\n? - Ayuda\nq - Cerrar aplicación\n\n"
        self.msg_help += self.msg_info

        self.err_cache = "Error borrando el caché"
        self.err_backend = "Parece ser que ningún programa para actualizar la imagen de fondo está instalado en su sistema.\n"
        self.err_backend += "Por favor, instala uno de los siguientes programas para poder cambiar la imagen de fondo:\n\n"
        self.err_backend += "- swaybg (Wayland)\n- swww (Wayland)\n"
        self.err_backend += "- hyprpaper (Wayland)\n- feh (Xorg)\n"
        self.err_backend += "- wallutils (Xorg & Wayland)\n- mpvpaper (Xorg & Wayland)\n\n"
        self.err_backend += self.msg_info

        self.tip_refresh = "Volver a almacenar la carpeta de imágenes"
        self.tip_fill = "Escoja el tipo de relleno"
        self.tip_backend = "Escoja el programa para cambiar la imagen de fondo"
        self.tip_sorting = "Escoja una forma de ordenar"
        self.tip_display = "Escoja el nombre de la pantalla"
        self.tip_color = "Escoja un color de fondo"
        self.tip_random = "Actualizar la imagen de fondo a una imagen aleatoria"
        self.tip_exit = "Cerrar la aplicación"
        self.tip_transition = "Elige el tipo de transición"
        self.tip_mpv_stop = "Detener todos los procesos de mpv"
        self.tip_mpv_pause = "Pausar fondo de pantalla de video"
        self.tip_mpv_sound = "Reproducir el sonido del video"
        self.tip_timer = "Frecuencia con la que se cambia automáticamente el fondo de pantalla"
        self.tip_start = "Iniciar el cambio automático del fondo de pantalla después de un intervalo de tiempo"


def load_language(lang):
    """Load the language package according to selected language"""
    if lang == "de":
        txt = German()
    elif lang == "fr":
        txt = French()
    elif lang == "ru":
        txt = Russian()
    elif lang == "by":
        txt = Belarusian()
    elif lang == "pl":
        txt = Polish()
    elif lang == "zh":
        txt = Chinese()
    elif lang == "zh_HK":
        txt = TraditionalChinese()
    elif lang == "es":
        txt = Spanish()
    else:
        txt = English()
    return txt




"""Module responsible for reading and saving the configuration and state files"""

import configparser
import pathlib
from argparse import Namespace
from typing import List
from platformdirs import user_config_path, user_pictures_path, user_cache_path, user_state_path

from waypaper.options import FILL_OPTIONS, SORT_OPTIONS, SWWW_TRANSITION_TYPES, BACKEND_OPTIONS
from waypaper.common import check_installed_backends


class Config:
    """User configuration loaded from the config.ini file"""
    def __init__(self):
        # All paths (folders or wallpapers) are Path objects
        self.name = "waypaper"
        self.home_path = pathlib.Path.home()
        self.image_folder_list: list[pathlib.Path] = []
        self.image_folder_fallback: str = str(user_pictures_path())
        self.installed_backends = check_installed_backends()
        self.selected_wallpaper = None
        self.selected_monitor = "All"
        self.fill_option = FILL_OPTIONS[0]
        self.sort_option = SORT_OPTIONS[0]
        self.backend = self.installed_backends[-1]
        self.color = "#ffffff"
        self.number_of_columns = 3
        self.swww_transition_type = SWWW_TRANSITION_TYPES[0]
        self.swww_transition_step = 90
        self.swww_transition_angle = 0
        self.swww_transition_duration = 2
        self.swww_transition_fps = 60
        self.mpvpaper_sound = False
        self.mpvpaper_options = ""
        self.lang = "en"
        self.monitors = [self.selected_monitor]
        self.wallpapers = []
        self.post_command = ""
        self.include_subfolders = False
        self.include_all_subfolders = False
        self.show_hidden = False
        self.show_gifs_only = False
        self.zen_mode = False
        self.cache_dir = user_cache_path(self.name)
        self.config_dir = user_config_path(self.name)
        self.config_file = self.config_dir / "config.ini"
        self.state_dir = user_state_path(self.name)
        self.state_file = self.state_dir / "state.ini"
        self.use_xdg_state = False
        self.use_post_command = True
        self.show_path_in_tooltip = True

        # Create config and cache folders:
        self.config_dir.mkdir(parents=True, exist_ok=True)
        self.cache_dir.mkdir(parents=True, exist_ok=True)
        self.state_dir.mkdir(parents=True, exist_ok=True)


    def select_wallpaper(self, path_str: str) -> None:
        self.selected_wallpaper = pathlib.Path(path_str)


    def shorten_path(self, path: pathlib.Path) -> str:
        """Replace home part of paths with tilde"""
        if path.is_relative_to(self.home_path):
            return str(path).replace(str(self.home_path), "~", 1)
        elif path:
            return str(path)
        else:
            return ''


    def get_image_folder_list(self, section: str, config) -> list[pathlib.Path]:
        image_folders_str: list[str] =  config.get(section, "folder", fallback = self.image_folder_fallback).split("\n")
        image_folder_list = [pathlib.Path(path_str).expanduser() for path_str in image_folders_str]
        return image_folder_list


    def read(self) -> None:
        """Load data from the config.ini or use default if it does not exists"""
        config = configparser.ConfigParser()
        config.read(self.config_file, 'utf-8')

        # Read basic parameters:
        self.fill_option = config.get("Settings", "fill", fallback=self.fill_option)
        self.sort_option = config.get("Settings", "sort", fallback=self.sort_option)
        self.backend = config.get("Settings", "backend", fallback=self.backend)
        self.color = config.get("Settings", "color", fallback=self.color)
        self.post_command = config.get("Settings", "post_command", fallback=self.post_command)
        self.swww_transition_type = config.get("Settings", "swww_transition_type", fallback=self.swww_transition_type)
        self.swww_transition_step = config.get("Settings", "swww_transition_step", fallback=self.swww_transition_step)
        self.swww_transition_angle = config.get("Settings", "swww_transition_angle", fallback=self.swww_transition_angle)
        self.swww_transition_duration = config.get("Settings", "swww_transition_duration", fallback=self.swww_transition_duration)
        self.swww_transition_fps = config.get("Settings", "swww_transition_fps", fallback=self.swww_transition_fps)
        self.mpvpaper_sound = config.getboolean("Settings", "mpvpaper_sound", fallback=self.mpvpaper_sound)
        self.mpvpaper_options = config.get("Settings", "mpvpaper_options", fallback=self.mpvpaper_options)
        self.number_of_columns = int(config.get("Settings", "number_of_columns", fallback=self.number_of_columns))
        self.lang = config.get("Settings", "language", fallback=self.lang)
        self.include_subfolders = config.getboolean("Settings", "subfolders", fallback=self.include_subfolders)
        self.include_all_subfolders = config.getboolean("Settings", "all_subfolders", fallback=self.include_all_subfolders)
        self.show_hidden = config.getboolean("Settings", "show_hidden", fallback=self.show_hidden)
        self.show_gifs_only = config.getboolean("Settings", "show_gifs_only", fallback=self.show_gifs_only)
        self.zen_mode = config.getboolean("Settings", "zen_mode", fallback=self.zen_mode)
        self.use_xdg_state = config.getboolean("Settings", "use_xdg_state", fallback=self.use_xdg_state)
        self.show_path_in_tooltip = config.getboolean("Settings", "show_path_in_tooltip", fallback=self.show_path_in_tooltip)

        # Read and convert strings representing lists and paths:
        monitors_str = config.get("Settings", "monitors", fallback=self.selected_monitor, raw=True)
        wallpapers_str = config.get("Settings", "wallpaper", fallback="", raw=True)
        self.image_folder_list = self.get_image_folder_list("Settings", config)
        if monitors_str:
            self.monitors = [str(monitor) for monitor in monitors_str.split(",")]
        if wallpapers_str:
            self.wallpapers = [pathlib.Path(paper).expanduser() for paper in wallpapers_str.split(",")]


    def read_state(self) -> None:
        """Load data from the state.ini file"""
        if not self.use_xdg_state:
            return

        state = configparser.ConfigParser()
        state.read(self.state_file, 'utf-8')

        # Read and convert strings representing lists and paths:
        monitors_str = state.get("State", "monitors", fallback=self.selected_monitor, raw=True)
        wallpapers_str = state.get("State", "wallpaper", fallback="", raw=True)
        self.image_folder_list = self.get_image_folder_list("State", state)
        if monitors_str:
            self.monitors = [str(monitor) for monitor in monitors_str.split(",")]
            self.selected_monitor = self.monitors[0]
        if wallpapers_str:
            self.wallpapers = [pathlib.Path(paper).expanduser() for paper in wallpapers_str.split(",")]


    def check_validity(self) -> None:
        """Check if the config parameters are valid and correct them if needed"""

        # Check validity of general options:
        if self.backend not in BACKEND_OPTIONS:
            self.backend = self.installed_backends[-1]
        if self.sort_option not in SORT_OPTIONS:
            self.sort_option = SORT_OPTIONS[0]
        if self.fill_option not in FILL_OPTIONS:
            self.fill_option = FILL_OPTIONS[0]
        if self.swww_transition_type not in SWWW_TRANSITION_TYPES:
            self.swww_transition_type = "any"
        if self.number_of_columns <= 0:
            self.number_of_columns = 1

        # Check validity of other swww options:
        if 0 > int(self.swww_transition_angle) > 180:
            self.swww_transition_angle = 0
        if 0 > int(self.swww_transition_step) > 255:
            self.swww_transition_step = 90
        if 0 > float(self.swww_transition_duration):
            self.swww_transition_duration = 2
        if 0 > int(self.swww_transition_fps):
            self.swww_transition_fps = 60


    def attribute_selected_wallpaper(self) -> None:
        """
        This function is used before saving config to attribute a wallpaper to a monitor.
        For example, if only certain monitor was affected,
        change only corresponding wallpaper in the wallpapers list.
        Otherwise, reset the list and set one wallpaper for all monitors.
        """
        if not self.selected_wallpaper:
            return
        if self.selected_monitor == "All":
            self.monitors = [self.selected_monitor]
            self.wallpapers = [self.selected_wallpaper]
        elif self.selected_monitor in self.monitors:
            index = self.monitors.index(self.selected_monitor)
            self.wallpapers[index] = self.selected_wallpaper
        else:
            self.monitors.append(self.selected_monitor)
            self.wallpapers.append(self.selected_wallpaper)


    def save_state_file(self) -> None:
        """Save some parameters to the state file"""
        state = configparser.ConfigParser()
        state.read(self.state_file)
        if not state.has_section("State"):
            state.add_section("State")
        self.write_folder_list_to_config("State", state)
        state.set("State", "monitors", ",".join(self.monitors))
        state.set("State", "wallpaper", ','.join(self.shorten_path(p) for p in self.wallpapers))
        try:
            with open(self.state_file, "w") as statefile:
                state.write(statefile)
        except PermissionError:
            print("Could not save state file due to permission error.")


    def write_folder_list_to_config(self, section: str, config):
        config_save_folder_str = ""
        for i, folder in enumerate(self.image_folder_list):
            if i == 0:
                config_save_folder_str += self.shorten_path(folder)
            else:
                config_save_folder_str += "\n     " + self.shorten_path(folder)
        config.set(section, "folder", config_save_folder_str)


    def save(self, only_state=False) -> None:
        """Save current parameters to the configuration file"""

        # Write configuration to the file:
        config = configparser.ConfigParser()
        config.read(self.config_file)
        if not config.has_section("Settings"):
            config.add_section("Settings")
        config.set("Settings", "language", self.lang)

        # If state file is used, some parameters are not save into config:
        if not self.use_xdg_state:
            self.write_folder_list_to_config("Settings", config)
            config.set("Settings", "monitors", ",".join(self.monitors))
            config.set("Settings", "wallpaper", ','.join(self.shorten_path(p) for p in self.wallpapers))

        # Save the parameters into config:
        config.set("Settings", "show_path_in_tooltip", str(self.show_path_in_tooltip))
        config.set("Settings", "backend", self.backend)
        config.set("Settings", "fill", self.fill_option)
        config.set("Settings", "sort", self.sort_option)
        config.set("Settings", "color", self.color)
        config.set("Settings", "subfolders", str(self.include_subfolders))
        config.set("Settings", "all_subfolders", str(self.include_all_subfolders))
        config.set("Settings", "show_hidden", str(self.show_hidden))
        config.set("Settings", "show_gifs_only", str(self.show_gifs_only))
        config.set("Settings", "zen_mode", str(self.zen_mode))
        config.set("Settings", "post_command", self.post_command)
        config.set("Settings", "number_of_columns", str(self.number_of_columns))
        config.set("Settings", "swww_transition_type", str(self.swww_transition_type))
        config.set("Settings", "swww_transition_step", str(self.swww_transition_step))
        config.set("Settings", "swww_transition_angle", str(self.swww_transition_angle))
        config.set("Settings", "swww_transition_duration", str(self.swww_transition_duration))
        config.set("Settings", "swww_transition_fps", str(self.swww_transition_fps))
        config.set("Settings", "mpvpaper_sound", str(self.mpvpaper_sound))
        config.set("Settings", "mpvpaper_options", str(self.mpvpaper_options))
        config.set("Settings", "use_xdg_state", str(self.use_xdg_state))

        try:
            with open(self.config_file, "w") as configfile:
                config.write(configfile)
        except PermissionError:
            print("Could not save config file due to permission error.")

        # If requested, save the state file:
        if self.use_xdg_state:
            self.save_state_file()


    def read_parameters_from_user_arguments(self, args: Namespace) -> None:
        """
        Read user arguments provided at the run. These values take priority over config.ini
        :param args Output of the argparse parser.parse_args() method. Should be only used in __main__.py.
        """
        if args.backend:
            self.backend = args.backend
        if args.fill:
            self.fill_option = args.fill
        if args.folder:
            self.image_folder_list = [pathlib.Path(path_str) for path_str in args.folder]
        if args.state_file:
            self.use_xdg_state = True # Use of a custom state file implies state is in a separate file, requires use_xdg_state
            self.state_file = pathlib.Path(args.state_file).expanduser()
        if args.no_post_command:
            self.use_post_command = False
